import os
import json
import xml.etree.ElementTree as ET
from xml.dom import minidom
from harnice import fileio, cli, state


default_desc = "TITLEBLOCK, PAPER SIZE, DESIGN"


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-params.json": "params",
        f"{state.partnumber('pn-rev')}-drawing.svg": "drawing",
        f"{state.partnumber('pn-rev')}-attributes.json": "attributes",
    }


def generate_structure():
    pass


def render():
    if (
        cli.prompt(
            "Warning: rendering a titleblock may clear user edits to its svg. Proceed?",
            default="yes",
        )
        != "yes"
    ):
        exit()

    # === Default Parameters ===
    params = {
        "page_size": [11 * 96, 8.5 * 96],
        "outer_margin": 20,
        "inner_margin": 40,
        "tick_spacing": 96,
        "tb_origin_offset": [398, 48],
        "row_heights": [24, 24],
        "column_widths": [[264, 50, 84], [73, 126, 139, 60]],
        "label_offset": [2, 7],
        "key_offset_y": 16,
        "cell_texts": [
            [
                ("DESCRIPTION", "tblock-key-desc"),
                ("REV", "tblock-key-rev"),
                ("PAGE DESC", "tblock-key-pagedesc"),
            ],
            [
                ("SCALE", "tblock-key-scale"),
                ("PART NUMBER", "tblock-key-pn"),
                ("DRAWN BY", "tblock-key-drawnby"),
                ("SHEET", "tblock-key-sheet"),
            ],
        ],
    }

    # === If param file doesn't exist, create it ===
    if not os.path.exists(fileio.path("params")):
        with open(fileio.path("params"), "w", encoding="utf-8") as f:
            json.dump(params, f, indent=2)
        # if it does exist, ignore it

    # === Load parameters from JSON ===
    with open(fileio.path("params"), "r", encoding="utf-8") as f:
        p = json.load(f)

    width, height = p["page_size"]
    svg = ET.Element(
        "svg",
        {
            "xmlns": "http://www.w3.org/2000/svg",
            "version": "1.1",
            "width": str(width),
            "height": str(height),
        },
    )

    contents_group = ET.SubElement(svg, "g", {"id": "tblock-contents-start"})

    def add_rect(parent, x, y, w, h, stroke="black", fill="none", stroke_width=1):
        ET.SubElement(
            parent,
            "rect",
            {
                "x": str(x),
                "y": str(y),
                "width": str(w),
                "height": str(h),
                "fill": fill,
                "stroke": stroke,
                "stroke-width": str(stroke_width),
            },
        )

    def add_text(parent, x, y, text, size=8, anchor="start", bold=False, id=None):
        style = f"font-size:{size}px;font-family:Arial"
        if bold:
            style += ";font-weight:bold"
        attrs = {
            "x": str(x),
            "y": str(y),
            "style": style,
            "text-anchor": anchor,
        }
        if id:
            attrs["id"] = id
        ET.SubElement(parent, "text", attrs).text = text

    # === Border Group ===
    border_group = ET.SubElement(contents_group, "g", {"id": "border"})

    x_ticks = int((width - 2 * p["inner_margin"]) // p["tick_spacing"])
    for i in range(x_ticks):
        x0 = p["inner_margin"] + i * p["tick_spacing"]
        x_center = x0 + p["tick_spacing"] / 2
        ET.SubElement(
            border_group,
            "line",
            {
                "x1": str(x0),
                "y1": str(p["outer_margin"]),
                "x2": str(x0),
                "y2": str(height - p["outer_margin"]),
                "stroke": "black",
                "stroke-width": "0.5",
            },
        )
        label_y_top = (p["outer_margin"] + p["inner_margin"]) / 2
        label_y_bot = height - label_y_top
        add_text(border_group, x_center, label_y_top, str(i + 1), anchor="middle")
        add_text(border_group, x_center, label_y_bot, str(i + 1), anchor="middle")

    x_end = p["inner_margin"] + x_ticks * p["tick_spacing"]
    ET.SubElement(
        border_group,
        "line",
        {
            "x1": str(x_end),
            "y1": str(p["outer_margin"]),
            "x2": str(x_end),
            "y2": str(height - p["outer_margin"]),
            "stroke": "black",
            "stroke-width": "0.5",
        },
    )

    y_ticks = int((height - 2 * p["inner_margin"]) // p["tick_spacing"])
    for j in range(y_ticks):
        y0 = p["inner_margin"] + j * p["tick_spacing"]
        y_center = y0 + p["tick_spacing"] / 2
        ET.SubElement(
            border_group,
            "line",
            {
                "x1": str(p["outer_margin"]),
                "y1": str(y0),
                "x2": str(width - p["outer_margin"]),
                "y2": str(y0),
                "stroke": "black",
                "stroke-width": "0.5",
            },
        )
        label = chr(ord("A") + j)
        label_x_left = (p["outer_margin"] + p["inner_margin"]) / 2
        label_x_right = width - label_x_left
        add_text(border_group, label_x_left, y_center + 4, label, anchor="middle")
        add_text(border_group, label_x_right, y_center + 4, label, anchor="middle")

    y_end = p["inner_margin"] + y_ticks * p["tick_spacing"]
    ET.SubElement(
        border_group,
        "line",
        {
            "x1": str(p["outer_margin"]),
            "y1": str(y_end),
            "x2": str(width - p["outer_margin"]),
            "y2": str(y_end),
            "stroke": "black",
            "stroke-width": "0.5",
        },
    )

    add_rect(
        border_group,
        p["outer_margin"],
        p["outer_margin"],
        width - 2 * p["outer_margin"],
        height - 2 * p["outer_margin"],
    )
    add_rect(
        border_group,
        p["inner_margin"],
        p["inner_margin"],
        width - 2 * p["inner_margin"],
        height - 2 * p["inner_margin"],
        stroke="black",
        fill="white",
        stroke_width=1,
    )

    # === Logo Group ===
    tb_origin_x = width - p["inner_margin"] - p["tb_origin_offset"][0]
    tb_origin_y = height - p["inner_margin"] - p["tb_origin_offset"][1]
    logo_width = 1.25 * 96
    logo_height = sum(p["row_heights"])
    logo_group = ET.SubElement(contents_group, "g", {"id": "logo"})
    add_rect(logo_group, tb_origin_x - logo_width, tb_origin_y, logo_width, logo_height)

    # === Titleblock Cell Groups ===
    y_cursor = tb_origin_y
    for row_idx, row_height in enumerate(p["row_heights"]):
        row_cols = p["column_widths"][row_idx]
        row_cells = p["cell_texts"][row_idx]
        x_cursor = tb_origin_x
        for col_idx, col_width in enumerate(row_cols):
            label, key_id = row_cells[col_idx]
            group_id = (
                label.lower().replace(" ", "-")
                if label
                else f"cell-r{row_idx}-c{col_idx}"
            )
            cell_group = ET.SubElement(contents_group, "g", {"id": group_id})
            add_rect(cell_group, x_cursor, y_cursor, col_width, row_height)

            if label:
                add_text(
                    cell_group,
                    x_cursor + p["label_offset"][0],
                    y_cursor + p["label_offset"][1],
                    label,
                    size=7,
                    bold=True,
                )
            if key_id:
                center_x = x_cursor + col_width / 2
                add_text(
                    cell_group,
                    center_x,
                    y_cursor + p["key_offset_y"],
                    key_id,
                    size=7,
                    anchor="middle",
                    id=key_id,
                )

            x_cursor += col_width
        y_cursor += row_height

    ET.SubElement(svg, "g", {"id": "tblock-contents-end"})
    rough_string = ET.tostring(svg, encoding="utf-8")
    pretty = minidom.parseString(rough_string).toprettyxml(indent="  ")
    with open(fileio.path("drawing"), "w", encoding="utf-8") as f:
        f.write(pretty)

    # === Write attributes file ===
    periphery_json = {
        "page_size_in": [
            round(p["page_size"][0] / 96, 3),
            round(p["page_size"][1] / 96, 3),
        ],
    }

    with open(fileio.path("attributes"), "w", encoding="utf-8") as f:
        json.dump(periphery_json, f, indent=2)

    print()
    print(f"Titleblock '{state.partnumber('pn')}' updated")
    print()



import re


# not initializing these variables so that a NameError is raised if they are not set
def set_pn(x):
    global pn
    pn = x


def set_rev(x):
    global rev
    rev = x

def set_product(x):
    global product
    product = x


def set_net(x):
    global net
    net = x


def set_file_structure(x):
    global file_structure
    file_structure = x


def partnumber(format):
    # Returns part numbers in various formats based on the current working directory

    # given a part number "pppppp-revR"

    # format options:
    # "pn-rev"    returns "pppppp-revR"
    # "pn"        returns "pppppp"
    # "rev"       returns "revR"
    # "R"         returns "R"

    pn_rev = f"{pn}-rev{rev}"

    if format == "pn-rev":
        return pn_rev

    elif format == "pn":
        match = re.search(r"-rev", pn_rev)
        if match:
            return pn_rev[: match.start()]

    elif format == "rev":
        match = re.search(r"-rev", pn_rev)
        if match:
            return pn_rev[match.start() + 1 :]

    elif format == "R":
        match = re.search(r"-rev", pn_rev)
        if match:
            return pn_rev[match.start() + 4 :]

    else:
        raise ValueError("Function 'partnumber' not presented with a valid format")



from harnice import fileio
from harnice.lists import instances_list
from harnice.utils import library_utils
import os


def end_ports_of_circuit(circuit_id):
    """
    Returns the instance names at the end ports (port 0 and maximum port) of a circuit.

    Finds and returns the instance names connected to port 0 and the maximum port number
    in the specified circuit. These represent the endpoints of the circuit.

    **Args:**
    - `circuit_id` (str): Circuit ID to look up. Must be a valid integer string.

    **Returns:**
    - `tuple`: A tuple of `(zero_port, max_port)` instance names. Either may be empty
        string if not found.

    **Raises:**
    - `ValueError`: If `circuit_id` is not a valid integer.
    """
    try:
        int(circuit_id)
    except ValueError:
        raise ValueError(f"Pass an integer circuit_id, not '{circuit_id}'")
    zero_port = ""
    max_port = ""
    for instance in fileio.read_tsv("instances list"):
        if instance.get("circuit_id") == circuit_id:
            if instance.get("circuit_port_number") == 0:
                zero_port = instance.get("instance_name")
            if instance.get("circuit_port_number") == max_port_number_in_circuit(
                circuit_id
            ):
                max_port = instance.get("instance_name")
    return zero_port, max_port


def max_port_number_in_circuit(circuit_id):
    """
    Finds the maximum circuit port number used in a circuit.

    Scans all instances in the circuit to find the highest port number assigned.
    Circuit instances (`item_type=="circuit"`) are skipped. Blank port numbers
    cause an error unless the instance is a circuit instance.

    **Args:**
    - `circuit_id` (str): Circuit ID to search.

    **Returns:**
    - `int`: The maximum port number found in the circuit (`0` if no ports found).

    **Raises:**
    - `ValueError`: If any non-circuit instance has a blank `circuit_port_number`.
    """
    max_port_number = 0
    for instance in fileio.read_tsv("instances list"):
        if instance.get("circuit_id") == circuit_id:
            if instance.get("circuit_port_number") == "":
                if instance.get("item_type") == "circuit":
                    continue
                raise ValueError(
                    f"Circuit port number is blank for {instance.get('instance_name')}"
                )
            max_port_number = max(
                max_port_number, int(instance.get("circuit_port_number"))
            )
    return max_port_number


def squeeze_instance_between_ports_in_circuit(
    instance_name, circuit_id, new_circuit_port_number
):
    """
    Inserts an instance into a circuit by shifting existing port numbers.

    Assigns the specified instance to a port number in the circuit, incrementing
    the port numbers of all instances that were at or after that port number.
    Circuit instances (`item_type=="circuit"`) are skipped and not renumbered.

    **Args:**
    - `instance_name` (str): Name of the instance to insert into the circuit.
    - `circuit_id` (str): Circuit ID to insert the instance into.
    - `new_circuit_port_number` (int): Port number to assign to the instance. All
        instances at this port number or higher will have their port numbers
        incremented by 1.
    """
    instances = fileio.read_tsv("instances list")
    for instance in instances:
        if instance.get("instance_name") == instance_name:
            continue
        if instance.get("circuit_id") == circuit_id:
            if instance.get("item_type") == "circuit":
                continue
            old_port_number = instance.get("circuit_port_number")
            if int(instance.get("circuit_port_number")) < new_circuit_port_number:
                continue
            else:
                instances_list.modify(
                    instance.get("instance_name"),
                    {"circuit_port_number": int(old_port_number) + 1},
                )
    for instance in instances:
        if instance.get("instance_name") == instance_name:
            instances_list.modify(
                instance_name,
                {
                    "circuit_id": circuit_id,
                    "circuit_port_number": new_circuit_port_number,
                },
            )


def instances_of_circuit(circuit_id):
    """
    Returns all instances in a circuit, sorted by port number.

    Finds all instances (excluding circuit instances themselves) that belong to
    the specified circuit and returns them sorted numerically by their `circuit_port_number`.
    Instances with missing port numbers are sorted last (treated as `999999`).

    **Args:**
    - `circuit_id` (str): Circuit ID to search for instances.

    **Returns:**
    - `list`: List of instance dictionaries, sorted by `circuit_port_number` in ascending order.
    """
    instances = []
    for instance in fileio.read_tsv("instances list"):
        if instance.get("circuit_id") == circuit_id:
            if instance.get("item_type") == "circuit":
                continue
            instances.append(instance)

    # sort numerically by circuit_port_number, treating missing as large number
    instances.sort(key=lambda x: int(x.get("circuit_port_number") or 999999))

    return instances


def instance_of_circuit_port_number(circuit_id, circuit_port_number):
    """
    Finds the instance name at a specific port number in a circuit.

    Searches the instances list for an instance that matches both the `circuit_id`
    and `circuit_port_number`. The comparison is done after stripping whitespace
    and converting to strings.

    **Args:**
    - `circuit_id` (str): Circuit ID to search.
    - `circuit_port_number` (str or int): Port number to search for.

    **Returns:**
    - `str`: The `instance_name` of the instance at the specified port number.

    **Raises:**
    - `ValueError`: If `circuit_id` or `circuit_port_number` is blank, or if no instance
        is found matching both the `circuit_id` and `circuit_port_number`.
    """
    if circuit_id in ["", None]:
        raise ValueError("Circuit ID is blank")
    if circuit_port_number in ["", None]:
        raise ValueError("Circuit port number is blank")

    for instance in fileio.read_tsv("instances list"):
        if instance.get("circuit_id").strip() == str(circuit_id).strip():
            if (
                instance.get("circuit_port_number").strip()
                == str(circuit_port_number).strip()
            ):
                return instance.get("instance_name")

    raise ValueError(
        f"No instance found for circuit {circuit_id} and port number {circuit_port_number}"
    )


def circuit_instance_of_instance(instance_name):
    """
    Returns the circuit instance dictionary for an instance that belongs to a circuit.

    Finds the circuit instance (`item_type=="circuit"`) that corresponds to a given
    instance. The circuit instance has the same `circuit_id` as the instance's `circuit_id`.

    **Args:**
    - `instance_name` (str): Name of the instance to find the circuit instance for.

    **Returns:**
    - `dict`: The circuit instance dictionary.

    **Raises:**
    - `ValueError`: If the circuit instance cannot be found for the given instance.
    """
    circuit_instance_name = ""
    instance_rows = fileio.read_tsv("instances list")
    for instance in instance_rows:
        if instance.get("instance_name") == instance_name:
            circuit_instance_name = instance.get("circuit_id")
            break
    for instance in instance_rows:
        if instance.get("circuit_id") == circuit_instance_name:
            if instance.get("instance_name") == instance_name:
                return instance
    raise ValueError(
        f"Circuit instance {circuit_instance_name} of instance {instance_name} not found"
    )


def assign_cable_conductor(
    cable_instance_name,  # unique identifier for the cable in your project
    cable_conductor_id,  # (container, identifier) tuple identifying the conductor in the cable being imported
    conductor_instance,  # instance name of the conductor in your project
    library_info,  # dict containing library info: {lib_repo, mpn, lib_subpath, used_rev}
    net,  # which net this cable belongs to
):
    """
    Assigns a conductor instance to a specific conductor in a cable.

    Links a conductor instance in the project to a specific conductor within a cable
    by importing the cable from the library (if not already imported) and updating
    the conductor instance with cable assignment information, including the conductor's
    appearance from the cable definition.

    The `cable_conductor_id` uses the `(container, identifier)` format from the cable
    conductor list. The cable is imported if it doesn't exist, and the conductor
    instance is updated with parent, group, container, identifier, and appearance info.

    **Args:**
    - `cable_instance_name` (str): Unique identifier for the cable in the project.
    - `cable_conductor_id` (tuple): Tuple of `(container, identifier)` identifying the
        conductor in the cable being imported.
    - `conductor_instance` (str): Instance name of the conductor in the project.
    - `library_info` (dict): Dictionary containing library information with keys:
        `lib_repo`, `mpn`, `lib_subpath`, and optionally `used_rev`.
    - `net` (str): Net name that this cable belongs to.

    **Raises:**
    - `ValueError`: If the conductor has already been assigned to another instance,
        or if the `conductor_instance` has already been assigned to another cable.
    """
    # for cable_conductor_id, see (container, identifier) from the cable conductor list.
    # TODO: ensure cable_conductor_id has the right format.

    instances = fileio.read_tsv("instances list")

    # --- Make sure conductor of cable has not been assigned yet
    for instance in instances:
        if instance.get("cable_group") == cable_instance_name:
            if instance.get("cable_container") == cable_conductor_id[0]:
                if instance.get("cable_identifier") == cable_conductor_id[1]:
                    raise ValueError(
                        f"when assingning '{cable_conductor_id} of '{cable_instance_name}' to '{conductor_instance}', "
                        f"conductor '{cable_conductor_id}' of '{cable_instance_name}' has already been assigned to {instance.get('instance_name')}"
                    )

    # --- Make sure conductor instance has not already been assigned to a cable
    for instance in instances:
        if instance.get("instance_name") == conductor_instance:
            if (
                instance.get("cable_group") not in ["", None]
                or instance.get("cable_container") not in ["", None]
                or instance.get("cable_identifier") not in ["", None]
            ):
                raise ValueError(
                    f"when assingning '{cable_conductor_id} of '{cable_instance_name}' to {conductor_instance}', "
                    f"instance '{conductor_instance}' has alredy been assigned to another cable"
                    f"to '{instance.get('cable_identifier')}' of cable '{instance.get('cable_group')}'"
                )

    cable_destination_directory = os.path.join(
        fileio.dirpath(None), "instance_data", "cable", cable_instance_name
    )

    instances_list.new_instance(
        cable_instance_name,
        {
            "net": net,
            "item_type": "cable",
            "location_type": "segment",
            "cable_group": cable_instance_name,
        },
        ignore_duplicates=True,
    )

    # --- Import cable from library ---
    library_utils.pull(
        {
            "lib_repo": library_info.get("lib_repo"),
            "lib_subpath": library_info.get("lib_subpath"),
            "item_type": "cable",
            "mpn": library_info.get("mpn"),
            "instance_name": cable_instance_name,
        },
        destination_directory=cable_destination_directory,
    )

    cable_attributes_path = os.path.join(
        cable_destination_directory, f"{cable_instance_name}-conductor_list.tsv"
    )
    cable_attributes = fileio.read_tsv(cable_attributes_path)

    # --- assign conductor
    for instance in instances:
        if instance.get("instance_name") == conductor_instance:
            appearance = None
            for row in cable_attributes:
                if (
                    row.get("container") == cable_conductor_id[0]
                    and row.get("identifier") == cable_conductor_id[1]
                ):
                    appearance = row.get("appearance")
                    break

            instances_list.modify(
                conductor_instance,
                {
                    "parent_instance": cable_instance_name,
                    "cable_group": cable_instance_name,
                    "cable_container": cable_conductor_id[0],
                    "cable_identifier": cable_conductor_id[1],
                    "appearance": appearance,
                },
            )
            break



import os
import json
import random
import math
import re
from PIL import Image, ImageDraw, ImageFont
from harnice import fileio, state
from harnice.utils import svg_utils


default_desc = "COTS COMPONENT, SIZE, COLOR, etc."


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-drawing.svg": "drawing",
        f"{state.partnumber('pn-rev')}-drawing.png": "drawing png",
        f"{state.partnumber('pn-rev')}-attributes.json": "attributes",
    }


def generate_structure():
    pass


def render():
    # === ATTRIBUTES JSON DEFAULTS ===
    default_attributes = {
        "tools": [],
        "build_notes": [],
        "csys_children": {
            "accessory-1": {"x": 3, "y": 2, "angle": 0, "rotation": 0},
            "accessory-2": {"x": 2, "y": 3, "angle": 0, "rotation": 0},
            "flagnote-1": {"angle": 0, "distance": 2, "rotation": 0},
            "flagnote-leader-1": {"angle": 0, "distance": 1, "rotation": 0},
            "flagnote-2": {"angle": 15, "distance": 2, "rotation": 0},
            "flagnote-leader-2": {"angle": 15, "distance": 1, "rotation": 0},
            "flagnote-3": {"angle": -15, "distance": 2, "rotation": 0},
            "flagnote-leader-3": {"angle": -15, "distance": 1, "rotation": 0},
            "flagnote-4": {"angle": 30, "distance": 2, "rotation": 0},
            "flagnote-leader-4": {"angle": 30, "distance": 1, "rotation": 0},
            "flagnote-5": {"angle": -30, "distance": 2, "rotation": 0},
            "flagnote-leader-5": {"angle": -30, "distance": 1, "rotation": 0},
            "flagnote-6": {"angle": 45, "distance": 2, "rotation": 0},
            "flagnote-leader-6": {"angle": 45, "distance": 1, "rotation": 0},
            "flagnote-7": {"angle": -45, "distance": 2, "rotation": 0},
            "flagnote-leader-7": {"angle": -45, "distance": 1, "rotation": 0},
            "flagnote-8": {"angle": 60, "distance": 2, "rotation": 0},
            "flagnote-leader-8": {"angle": 60, "distance": 1, "rotation": 0},
            "flagnote-9": {"angle": -60, "distance": 2, "rotation": 0},
            "flagnote-leader-9": {"angle": -60, "distance": 1, "rotation": 0},
            "flagnote-10": {"angle": -75, "distance": 2, "rotation": 0},
            "flagnote-leader-10": {"angle": -75, "distance": 1, "rotation": 0},
            "flagnote-11": {"angle": 75, "distance": 2, "rotation": 0},
            "flagnote-leader-11": {"angle": 75, "distance": 1, "rotation": 0},
            "flagnote-12": {"angle": -90, "distance": 2, "rotation": 0},
            "flagnote-leader-12": {"angle": -90, "distance": 1, "rotation": 0},
            "flagnote-13": {"angle": 90, "distance": 2, "rotation": 0},
            "flagnote-leader-13": {"angle": 90, "distance": 1, "rotation": 0},
            "flagnote-14": {"angle": -105, "distance": 2, "rotation": 0},
            "flagnote-leader-14": {"angle": -105, "distance": 1, "rotation": 0},
            "flagnote-15": {"angle": 105, "distance": 2, "rotation": 0},
            "flagnote-leader-15": {"angle": 105, "distance": 1, "rotation": 0},
            "flagnote-16": {"angle": -120, "distance": 2, "rotation": 0},
            "flagnote-leader-16": {"angle": -120, "distance": 1, "rotation": 0},
        },
    }

    attributes_path = fileio.path("attributes")

    # Load or create attributes.json
    if os.path.exists(attributes_path):
        try:
            with open(attributes_path, "r", encoding="utf-8") as f:
                attrs = json.load(f)
        except Exception as e:
            print(f"[WARNING] Could not load existing attributes.json: {e}")
            attrs = default_attributes.copy()
    else:
        attrs = default_attributes.copy()
        with open(attributes_path, "w", encoding="utf-8") as f:
            json.dump(attrs, f, indent=4)

    # === SVG SETUP ===
    svg_path = fileio.path("drawing")
    temp_svg_path = svg_path + ".tmp"

    svg_width = 400
    svg_height = 400
    group_name = f"{state.partnumber('pn')}-drawing"

    random_fill = "#{:06X}".format(random.randint(0, 0xFFFFFF))
    fallback_rect = f'    <rect x="0" y="-48" width="96" height="96" fill="{random_fill}" stroke="black" stroke-width="1"/>'

    csys_children = attrs.get("csys_children", {})

    lines = [
        '<?xml version="1.0" encoding="UTF-8"?>',
        f'<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="{svg_width}" height="{svg_height}">',
        f'  <g id="{group_name}-contents-start">',
        fallback_rect,
        "  </g>",
        f'  <g id="{group_name}-contents-end">',
        "  </g>",
    ]

    # === Render Output Csys Locations ===
    lines.append('  <g id="output csys locations">')

    arrow_len = 24
    dot_radius = 4
    arrow_size = 6

    for csys_name, csys in csys_children.items():
        try:
            x = float(csys.get("x", 0)) * 96
            y = float(csys.get("y", 0)) * 96

            angle_deg = float(csys.get("angle", 0))
            distance_in = float(csys.get("distance", 0))
            angle_rad = math.radians(angle_deg)
            dist_px = distance_in * 96
            x += dist_px * math.cos(angle_rad)
            y += dist_px * math.sin(angle_rad)

            rotation_deg = float(csys.get("rotation", 0))
            rotation_rad = math.radians(rotation_deg)
            cos_r, sin_r = math.cos(rotation_rad), math.sin(rotation_rad)

            dx_x, dy_x = arrow_len * cos_r, arrow_len * sin_r
            dx_y, dy_y = -arrow_len * sin_r, arrow_len * cos_r

            lines.append(f'    <g id="{csys_name}">')
            lines.append(
                f'      <circle cx="{x:.2f}" cy="{-y:.2f}" r="{dot_radius}" fill="black"/>'
            )

            def draw_arrow(x1, y1, dx, dy, color):
                x2, y2 = x1 + dx, y1 + dy
                lines.append(
                    f'      <line x1="{x1:.2f}" y1="{-y1:.2f}" '
                    f'x2="{x2:.2f}" y2="{-y2:.2f}" stroke="{color}" stroke-width="2"/>'
                )
                length = math.hypot(dx, dy)
                if length == 0:
                    return
                ux, uy = dx / length, dy / length
                px, py = -uy, ux
                base_x = x2 - ux * arrow_size
                base_y = y2 - uy * arrow_size
                tip = (x2, y2)
                left = (base_x + px * (arrow_size / 2), base_y + py * (arrow_size / 2))
                right = (base_x - px * (arrow_size / 2), base_y - py * (arrow_size / 2))
                lines.append(
                    f'      <polygon points="{tip[0]:.2f},{-tip[1]:.2f} '
                    f'{left[0]:.2f},{-left[1]:.2f} {right[0]:.2f},{-right[1]:.2f}" fill="{color}"/>'
                )

            draw_arrow(x, y, dx_x, dy_x, "red")
            draw_arrow(x, y, dx_y, dy_y, "green")
            lines.append("    </g>")

        except Exception as e:
            print(f"[WARNING] Failed to render csys {csys_name}: {e}")

    lines.append("  </g>")
    lines.append("</svg>")

    with open(temp_svg_path, "w", encoding="utf-8") as f:
        f.write("\n".join(lines) + "\n")

    if os.path.exists(svg_path):
        try:
            with open(svg_path, "r", encoding="utf-8") as f:
                svg_text = f.read()
        except Exception:
            svg_text = ""

        if (
            f"{group_name}-contents-start" not in svg_text
            or f"{group_name}-contents-end" not in svg_text
        ):
            svg_utils.add_entire_svg_file_contents_to_group(svg_path, group_name)

        svg_utils.find_and_replace_svg_group(
            source_svg_filepath=svg_path,
            source_group_name=group_name,
            destination_svg_filepath=temp_svg_path,
            destination_group_name=group_name,
        )

    if os.path.exists(svg_path):
        os.remove(svg_path)
    os.rename(temp_svg_path, svg_path)

    # ==================================================
    # PNG generation (SVG = truth for graphics; JSON = truth for CSYS)
    # ==================================================

    # ------------------------------------------------------------------
    # 1. Extract raw contents group from final SVG
    # ------------------------------------------------------------------
    with open(svg_path, "r", encoding="utf-8") as f:
        svg_text = f.read()

    start_tag = f'<g id="{group_name}-contents-start">'
    end_tag = f'<g id="{group_name}-contents-end">'

    start_idx = svg_text.find(start_tag)
    end_idx = svg_text.find(end_tag)

    if start_idx == -1 or end_idx == -1:
        print(
            "[WARNING] Could not find contents group in SVG — PNG will only draw csys."
        )
        inner_svg = ""
    else:
        inner_svg = svg_text[start_idx + len(start_tag) : end_idx]

    # ======================================================
    # 2. Utility functions + transform matrix code
    # ======================================================

    def get_attr(tag, name, default=None):
        m = re.search(rf'{name}="([^"]+)"', tag)
        return m.group(1) if m else default

    def _parse_floats(s):
        return [float(x) for x in re.split(r"[ ,]+", s.strip()) if x]

    def _mat_identity():
        return [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]

    def _mat_mul(a, b):
        res = [[0.0, 0.0, 0.0] for _ in range(3)]
        for i in range(3):
            for j in range(3):
                res[i][j] = a[i][0] * b[0][j] + a[i][1] * b[1][j] + a[i][2] * b[2][j]
        return res

    def _mat_apply(m, x, y):
        nx = m[0][0] * x + m[0][1] * y + m[0][2]
        ny = m[1][0] * x + m[1][1] * y + m[1][2]
        return nx, ny

    def parse_transform(transform_str):
        """Parse SVG transform into 3x3 matrix."""
        if not transform_str:
            return _mat_identity()

        mat = _mat_identity()
        items = re.findall(
            r"(matrix|translate|scale|rotate|skewX|skewY)\s*\(([^)]*)\)", transform_str
        )
        for op, args_str in items:
            nums = _parse_floats(args_str)
            op = op.lower()

            if op == "matrix" and len(nums) == 6:
                a, b, c, d, e, f = nums
                m2 = [
                    [a, c, e],
                    [b, d, f],
                    [0.0, 0.0, 1.0],
                ]
            elif op == "translate":
                tx = nums[0] if len(nums) else 0.0
                ty = nums[1] if len(nums) > 1 else 0.0
                m2 = [
                    [1.0, 0.0, tx],
                    [0.0, 1.0, ty],
                    [0.0, 0.0, 1.0],
                ]
            elif op == "scale":
                sx = nums[0] if len(nums) else 1.0
                sy = nums[1] if len(nums) > 1 else sx
                m2 = [
                    [sx, 0.0, 0.0],
                    [0.0, sy, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            elif op == "rotate":
                angle = nums[0] if nums else 0.0
                rad = math.radians(angle)
                cos_a, sin_a = math.cos(rad), math.sin(rad)

                if len(nums) == 3:
                    cx, cy = nums[1], nums[2]
                    t1 = [[1, 0, cx], [0, 1, cy], [0, 0, 1]]
                    r = [[cos_a, -sin_a, 0], [sin_a, cos_a, 0], [0, 0, 1]]
                    t2 = [[1, 0, -cx], [0, 1, -cy], [0, 0, 1]]
                    m2 = _mat_mul(_mat_mul(t1, r), t2)
                else:
                    m2 = [
                        [cos_a, -sin_a, 0.0],
                        [sin_a, cos_a, 0.0],
                        [0.0, 0.0, 1.0],
                    ]
            elif op == "skewx":
                angle = nums[0] if nums else 0.0
                t = math.tan(math.radians(angle))
                m2 = [[1, t, 0], [0, 1, 0], [0, 0, 1]]
            elif op == "skewy":
                angle = nums[0] if nums else 0.0
                t = math.tan(math.radians(angle))
                m2 = [[1, 0, 0], [t, 1, 0], [0, 0, 1]]
            else:
                m2 = _mat_identity()

            mat = _mat_mul(mat, m2)

        return mat

    # ======================================================
    # 3. Parse actual shapes into SVG pixel space
    # ======================================================

    parsed_shapes = []

    # -------- RECTANGLE --------
    for tag in re.findall(r"<rect[^>]*/?>", inner_svg):
        x = float(get_attr(tag, "x", 0))
        y = float(get_attr(tag, "y", 0))
        w = float(get_attr(tag, "width", 0))
        h = float(get_attr(tag, "height", 0))
        fill = get_attr(tag, "fill", "none")
        stroke = get_attr(tag, "stroke", None)
        stroke_w = float(get_attr(tag, "stroke-width", 1) or 1)
        transform_str = get_attr(tag, "transform", "")

        if transform_str:
            mat = parse_transform(transform_str)
            pts = [
                (x, y),
                (x + w, y),
                (x + w, y + h),
                (x, y + h),
            ]
            pts_tr = [_mat_apply(mat, px, py) for px, py in pts]
            parsed_shapes.append(
                (
                    "polygon",
                    {
                        "points": pts_tr,
                        "fill": fill,
                        "stroke": stroke,
                        "sw": stroke_w,
                    },
                )
            )
        else:
            parsed_shapes.append(
                (
                    "rect",
                    {
                        "x": x,
                        "y": y,
                        "w": w,
                        "h": h,
                        "fill": fill,
                        "stroke": stroke,
                        "sw": stroke_w,
                    },
                )
            )

    # -------- CIRCLE --------
    for tag in re.findall(r"<circle[^>]*/?>", inner_svg):
        cx = float(get_attr(tag, "cx", 0))
        cy = float(get_attr(tag, "cy", 0))
        r = float(get_attr(tag, "r", 0))
        fill = get_attr(tag, "fill", "none")
        stroke = get_attr(tag, "stroke", None)
        stroke_w = float(get_attr(tag, "stroke-width", 1) or 1)
        transform_str = get_attr(tag, "transform", "")

        if transform_str:
            mat = parse_transform(transform_str)
            cx, cy = _mat_apply(mat, cx, cy)

        parsed_shapes.append(
            (
                "circle",
                {
                    "cx": cx,
                    "cy": cy,
                    "r": r,
                    "fill": fill,
                    "stroke": stroke,
                    "sw": stroke_w,
                },
            )
        )

    # -------- LINE --------
    for tag in re.findall(r"<line[^>]*/?>", inner_svg):
        x1 = float(get_attr(tag, "x1", 0))
        y1 = float(get_attr(tag, "y1", 0))
        x2 = float(get_attr(tag, "x2", 0))
        y2 = float(get_attr(tag, "y2", 0))
        stroke = get_attr(tag, "stroke", "black")
        stroke_w = float(get_attr(tag, "stroke-width", 1) or 1)
        transform_str = get_attr(tag, "transform", "")

        if transform_str:
            mat = parse_transform(transform_str)
            x1, y1 = _mat_apply(mat, x1, y1)
            x2, y2 = _mat_apply(mat, x2, y2)

        parsed_shapes.append(
            (
                "line",
                {
                    "x1": x1,
                    "y1": y1,
                    "x2": x2,
                    "y2": y2,
                    "stroke": stroke,
                    "sw": stroke_w,
                },
            )
        )

    # -------- POLYGON --------
    for tag in re.findall(r"<polygon[^>]*/?>", inner_svg):
        pts_raw = get_attr(tag, "points", "")
        pts = []
        for p in pts_raw.split():
            if "," in p:
                xx, yy = p.split(",")
                pts.append((float(xx), float(yy)))
        fill = get_attr(tag, "fill", "none")
        stroke = get_attr(tag, "stroke", None)
        stroke_w = float(get_attr(tag, "stroke-width", 1) or 1)
        transform_str = get_attr(tag, "transform", "")

        if transform_str:
            mat = parse_transform(transform_str)
            pts = [_mat_apply(mat, px, py) for px, py in pts]

        parsed_shapes.append(
            ("polygon", {"points": pts, "fill": fill, "stroke": stroke, "sw": stroke_w})
        )

    # -------- TEXT (simple) --------
    for full_tag in re.findall(r"<text[^>]*>.*?</text>", inner_svg, flags=re.DOTALL):
        txt = re.sub(r"<.*?>", "", full_tag)
        x = float(get_attr(full_tag, "x", 0))
        y = float(get_attr(full_tag, "y", 0))
        fill = get_attr(full_tag, "fill", "black")
        transform_str = get_attr(full_tag, "transform", "")

        if transform_str:
            mat = parse_transform(transform_str)
            x, y = _mat_apply(mat, x, y)

        parsed_shapes.append(
            ("text", {"x": x, "y": y, "text": txt.strip(), "fill": fill})
        )

    # ======================================================
    # 4. CSYS → SVG pixel coordinates
    # ======================================================

    padding = 50
    scale = 96  # px per inch
    arrow_len_svg = 24

    def csys_svg_xy(csys):
        raw_x = csys.get("x")
        raw_y = csys.get("y")
        raw_d = csys.get("distance")
        raw_a = csys.get("angle")

        if raw_x not in ("", None) and raw_y not in ("", None):
            x_in = float(raw_x)
            y_in = float(raw_y)
        elif raw_d not in ("", None) and raw_a not in ("", None):
            dist = float(raw_d)
            ang = math.radians(float(raw_a))
            x_in = dist * math.cos(ang)
            y_in = dist * math.sin(ang)
        else:
            x_in, y_in = 0.0, 0.0

        # inches → SVG px ; y-up → y-down
        return x_in * scale, -y_in * scale

    def csys_svg_axes_endpoints(csys, base_x, base_y, arrow_len):
        rot = math.radians(float(csys.get("rotation", 0) or 0))

        # CSYS-space directions (Y-up)
        dx_x = math.cos(rot)
        dy_x = math.sin(rot)
        dx_y = -math.sin(rot)
        dy_y = math.cos(rot)

        # Convert CSYS → SVG (flip y)
        dx_x_s = dx_x
        dy_x_s = -dy_x
        dx_y_s = dx_y
        dy_y_s = -dy_y

        # Endpoints in SVG space
        x_x = base_x + dx_x_s * arrow_len
        y_x = base_y + dy_x_s * arrow_len
        x_y = base_x + dx_y_s * arrow_len
        y_y = base_y + dy_y_s * arrow_len

        return (x_x, y_x), (x_y, y_y)

    # ======================================================
    # 5. Compute bounding box from (SVG shapes + CSYS)
    # ======================================================

    pts = []

    # SVG shapes
    for typ, p in parsed_shapes:
        if typ == "rect":
            pts += [(p["x"], p["y"]), (p["x"] + p["w"], p["y"] + p["h"])]
        elif typ == "circle":
            pts += [
                (p["cx"] - p["r"], p["cy"] - p["r"]),
                (p["cx"] + p["r"], p["cy"] + p["r"]),
            ]
        elif typ == "line":
            pts += [(p["x1"], p["y1"]), (p["x2"], p["y2"])]
        elif typ == "polygon":
            pts += p["points"]
        elif typ == "text":
            pts.append((p["x"], p["y"]))

    # CSYS
    for csys_name, csys in csys_children.items():
        bx, by = csys_svg_xy(csys)
        pts.append((bx, by))
        (x_x, y_x), (x_y, y_y) = csys_svg_axes_endpoints(csys, bx, by, arrow_len_svg)
        pts += [(x_x, y_x), (x_y, y_y)]

    if not pts:
        pts = [(0, 0)]

    xs = [p[0] for p in pts]
    ys = [p[1] for p in pts]

    min_x, max_x = min(xs), max(xs)
    min_y, max_y = min(ys), max(ys)

    width = int((max_x - min_x) + 2 * padding)
    height = int((max_y - min_y) + 2 * padding)

    def map_xy(x, y):
        return (
            int((x - min_x) + padding),
            int((y - min_y) + padding),
        )

    # ======================================================
    # 6. Create PNG canvas and draw everything
    # ======================================================

    img = Image.new("RGB", (width, height), "white")
    draw = ImageDraw.Draw(img)

    try:
        font = ImageFont.truetype("Arial.ttf", 8)
    except Exception:
        font = ImageFont.load_default()

    # --- SHAPES ---
    for typ, p in parsed_shapes:
        if typ == "rect":
            x1, y1 = map_xy(p["x"], p["y"])
            x2, y2 = map_xy(p["x"] + p["w"], p["y"] + p["h"])
            draw.rectangle(
                (x1, y1, x2, y2),
                fill=p["fill"],
                outline=p["stroke"],
                width=int(p["sw"]),
            )
        elif typ == "circle":
            cx, cy = map_xy(p["cx"], p["cy"])
            r = p["r"]
            draw.ellipse(
                (cx - r, cy - r, cx + r, cy + r),
                fill=p["fill"],
                outline=p["stroke"],
                width=int(p["sw"]),
            )
        elif typ == "line":
            x1, y1 = map_xy(p["x1"], p["y1"])
            x2, y2 = map_xy(p["x2"], p["y2"])
            draw.line((x1, y1, x2, y2), fill=p["stroke"], width=int(p["sw"]))
        elif typ == "polygon":
            pts2 = [map_xy(x, y) for x, y in p["points"]]
            draw.polygon(pts2, fill=p["fill"], outline=p["stroke"])
        elif typ == "text":
            tx, ty = map_xy(p["x"], p["y"])
            draw.text((tx, ty), p["text"], fill=p["fill"], font=font)

    # --- CSYS ---
    dot_radius = 4

    for csys_name, csys in csys_children.items():
        bx, by = csys_svg_xy(csys)
        cx, cy = map_xy(bx, by)

        # Dot
        draw.ellipse(
            (cx - dot_radius, cy - dot_radius, cx + dot_radius, cy + dot_radius),
            fill="black",
        )

        # Endpoints in SVG → map to PNG
        (x_x, y_x), (x_y, y_y) = csys_svg_axes_endpoints(csys, bx, by, arrow_len_svg)
        px1, py1 = map_xy(x_x, y_x)
        px2, py2 = map_xy(x_y, y_y)

        # Axes
        draw.line((cx, cy, px1, py1), fill="red", width=2)  # X
        draw.line((cx, cy, px2, py2), fill="green", width=2)  # Y

        # Label
        draw.text((cx + 6, cy - 6), csys_name, fill="blue", font=font)

    # ======================================================
    # 7. Save PNG
    # ======================================================

    png_path = fileio.path("drawing png")
    img.save(png_path, dpi=(1000, 1000))

    print(f"Part file '{state.partnumber('pn')}' updated\n")



import os
import json
import random
import math
from harnice import fileio, cli, state


default_desc = "FLAGNOTE, PURPOSE"


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-params.json": "params",
        f"{state.partnumber('pn-rev')}-drawing.svg": "drawing",
    }


def generate_structure():
    pass


def render():
    if (
        cli.prompt(
            "Warning: rendering a titleblock may clear user edits to its svg. Proceed?",
            default="yes",
        )
        != "yes"
    ):
        exit()

    # Geometry generators
    def regular_ngon(n, radius=19.2, rotation_deg=0):
        angle_offset = math.radians(rotation_deg)
        return [
            [
                round(radius * math.cos(2 * math.pi * i / n + angle_offset), 2),
                round(radius * math.sin(2 * math.pi * i / n + angle_offset), 2),
            ]
            for i in range(n)
        ]

    def right_arrow():
        return [[-24, -12], [0, -12], [0, -24], [24, 0], [0, 24], [0, 12], [-24, 12]]

    def left_arrow():
        return [[24, -12], [0, -12], [0, -24], [-24, 0], [0, 24], [0, 12], [24, 12]]

    def flag_pennant():
        return [[-24, -12], [24, 0], [-24, 12]]

    # List of shape options with (label, generator)
    shape_options = [
        ("circle", None),
        ("square", lambda: regular_ngon(4, rotation_deg=45)),
        ("triangle", lambda: regular_ngon(3, rotation_deg=-90)),
        ("upside down triangle", lambda: regular_ngon(3, rotation_deg=90)),
        ("hexagon", lambda: regular_ngon(6)),
        ("pentagon", lambda: regular_ngon(5)),
        ("right arrow", right_arrow),
        ("left arrow", left_arrow),
        ("octagon", lambda: regular_ngon(8)),
        ("diamond", lambda: regular_ngon(4, rotation_deg=0)),
        ("flag / pennant", flag_pennant),
    ]

    # === Prompt shape if no params exist ===
    if not os.path.exists(fileio.path("params")):
        print("No flagnote params found.")
        print("Choose a shape for your flagnote:")
        for i, (label, _) in enumerate(shape_options, 1):
            print(f"  {i}) {label}")

        while True:
            response = cli.prompt("Enter the number of your choice").strip()
            if response.isdigit():
                index = int(response)
                if 1 <= index <= len(shape_options):
                    shape_label, shape_func = shape_options[index - 1]
                    break
            print("Invalid selection. Please enter a number from the list.")

        params = {"fill": 0xFFFFFF, "border": 0x000000, "text inside": "flagnote-text"}

        if shape_func:
            params["vertices"] = shape_func()

        with open(fileio.path("params"), "w", encoding="utf-8") as f:
            json.dump(params, f, indent=2)

    # === Load params ===
    with open(fileio.path("params"), "r", encoding="utf-8") as f:
        p = json.load(f)

    svg_width = 6 * 96
    svg_height = 6 * 96

    fill = p.get("fill")
    if not isinstance(fill, int):
        fill = random.randint(0x000000, 0xFFFFFF)

    border = p.get("border", 0x000000)
    shape_svg = ""

    # === Shape element ===
    if "vertices" in p:
        if p["vertices"]:
            points_str = " ".join(f"{x},{y}" for x, y in p["vertices"])
            shape_svg = f'    <polygon points="{points_str}" fill="#{fill:06X}" stroke="#{border:06X}"/>\n'
    else:
        shape_svg = f'    <circle cx="0" cy="0" r="10" fill="#{fill:06X}" stroke="#{border:06X}"/>\n'

    # === Text element ===
    text_content = p.get("text inside", "")
    text_svg = (
        f'    <text x="0" y="0" '
        f'style="font-size:8px;font-family:Arial" '
        f'text-anchor="middle" dominant-baseline="middle" id="flagnote-text">{text_content}</text>\n'
    )

    contents = shape_svg + text_svg if shape_svg else ""

    lines = [
        '<?xml version="1.0" encoding="UTF-8"?>',
        f'<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width="{svg_width}" height="{svg_height}">',
        f'  <g id="{state.partnumber("pn")}-drawing-contents-start">',
        contents.rstrip(),
        "  </g>",
        f'  <g id="{state.partnumber("pn")}-drawing-contents-end">',
        "  </g>",
        "</svg>",
    ]

    with open(fileio.path("drawing"), "w", encoding="utf-8") as f:
        f.write("\n".join(lines) + "\n")

    print()
    print(f"Flagnote '{state.partnumber('pn')}' updated")
    print()



import runpy
import os
from harnice import fileio, cli, state
from harnice.lists import instances_list, library_history

default_desc = "HARNESS, DOES A, FOR B"


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-feature_tree.py": "feature tree",
        f"{state.partnumber('pn-rev')}-instances_list.tsv": "instances list",
        f"{state.partnumber('pn-rev')}-formboard_graph_definition.png": "formboard graph definition png",
        f"{state.partnumber('pn-rev')}-library_import_history.tsv": "library history",
        "interactive_files": {
            f"{state.partnumber('pn-rev')}.formboard_graph_definition.tsv": "formboard graph definition",
        },
    }


def generate_structure():
    os.makedirs(
        fileio.dirpath("interactive_files", structure_dict=file_structure()),
        exist_ok=True,
    )


def render():
    """
    Main harness rendering entrypoint.
    """

    feature_tree_path = fileio.path("feature tree", structure_dict=file_structure())

    # ======================================================================
    # 1. Feature tree does not exist: prompt user and create it
    # ======================================================================
    if not os.path.exists(feature_tree_path):
        # ------------------------------------------------------------------
        # Ask whether this harness pulls from a system or not
        # ------------------------------------------------------------------
        print(
            "  's'   Enter 's' for system (or just hit enter) "
            "if this harness is pulling data from a system instances list"
        )
        print(
            "  'n'   Enter 'n' for none to build your harness entirely "
            "out of rules in feature tree (you're hardcore)"
        )
        build_macro = cli.prompt("")

        # ------------------------------------------------------------------
        # If harness is built from a system, ask for system parameters
        # ------------------------------------------------------------------
        if build_macro in (None, "", "s"):
            system_pn = cli.prompt("Enter the system part number")
            system_rev = cli.prompt("Enter the system revision id (ex. rev1)")
            target_net = cli.prompt("Enter the net you want to build this harness from")

            # Inject actual values into template
            build_macro_block = default_build_macro_block(
                system_pn, system_rev, target_net
            )
            push_block = default_push_block()

        # ------------------------------------------------------------------
        # Hardcore mode — no system importing
        # ------------------------------------------------------------------
        elif build_macro == "n":
            build_macro_block = ""
            push_block = ""

        else:
            print(
                "Unrecognized input. If you meant to select a template not listed, "
                "just select a template, delete the contents and start over manually. rip."
            )
            exit()

        # ------------------------------------------------------------------
        # Write feature tree file
        # ------------------------------------------------------------------
        feature_tree_text = _make_feature_tree(build_macro_block, push_block)

        with open(feature_tree_path, "w", encoding="utf-8") as f:
            f.write(feature_tree_text)

    # ======================================================================
    # 2. Init library + instances list
    # ======================================================================
    library_history.new()
    instances_list.new()
    instances_list.new_instance(
        "origin",
        {
            "instance_name": "origin",
            "item_type": "origin",
            "location_type": "node",
        },
    )

    # ======================================================================
    # 3. Run the feature tree
    # ======================================================================
    cli.print_import_status_headers()
    runpy.run_path(feature_tree_path, run_name="__main__")

    print(f"Harnice: harness {state.partnumber('pn')} rendered successfully!\n")


# ======================================================================
# Helper: prompt for system/no-system
# ======================================================================
def _prompt_build_macro():
    print(
        "  's'   Enter 's' for system (or just hit enter) "
        "if this harness is pulling data from a system instances list"
    )
    print(
        "  'n'   Enter 'n' for none to build your harness entirely "
        "out of rules in feature tree (you're hardcore)"
    )
    return cli.prompt("")


# ======================================================================
# Helper: build the complete feature tree file text
# ======================================================================
def _make_feature_tree(build_macro_block: str, push_block: str) -> str:
    """
    Insert build_macro_block and push_block into your exact template.
    """

    return f"""
import os
from harnice import fileio, state
from harnice.utils import (
    circuit_utils,
    formboard_utils,
    note_utils,
    library_utils,
    feature_tree_utils,
)
from harnice.lists import (
    instances_list,
    post_harness_instances_list,
    rev_history,
)

{build_macro_block}

# ===========================================================================
#                  HARNESS BUILD RULES
# ===========================================================================
# example: assign a special contact to one specific conductor
instances = fileio.read_tsv("instances list")
circuit_instance = None
connector_at_end_a = None
for instance in instances:
    if instance.get("channel_group") == "channel-MIC2.out1-PREAMP1.in2":
        if instance.get("signal_of_channel_type") == "pos":
            if instance.get("item_type") == "circuit":
                circuit_instance = instance
                connector_at_end_a = instances_list.attribute_of(instance.get("node_at_end_a"), "connector_group")
new_instance_name = f"{{circuit_instance.get('instance_name')}}-special_contact"
circuit_id = int(circuit_instance.get("circuit_id"))
instances_list.new_instance(
    new_instance_name, {{
        "bom_line_number": True,
        "mpn": "TXPS20",
        "item_type": "contact",
        "location_type": "node",
        "circuit_id": circuit_id,
        "connector_group": connector_at_end_a
    }}
)
circuit_utils.squeeze_instance_between_ports_in_circuit(
    new_instance_name, circuit_id, 1
)

# example: add a backshell
for instance in instances:
    if instance.get("instance_name") in ["X1.B.conn", "PREAMP2.in2.conn"]:
        instances_list.new_instance(f"{{instance.get('connector_group')}}.bs", {{
            "bom_line_number": True,
            "mpn": "M85049-90_9Z03",
            "item_type": "backshell",
            "parent_instance": instance.get("instance_name"),
            "location_type": "node",
            "connector_group": instance.get("connector_group"),
            "parent_csys_instance_name": (instances_list.instance_in_connector_group_with_item_type(instance.get("connector_group"), "node")).get("instance_name"),
            "parent_csys_outputcsys_name": "origin",
            "lib_repo": "https://github.com/harnice/harnice"
        }})
        instances_list.modify(instance.get("instance_name"), {{
            "parent_csys_instance_name": f"{{instance.get('connector_group')}}.bs",
            "parent_csys_outputcsys_name": "connector",
        }})


# ===========================================================================
#                   IMPORT PARTS FROM LIBRARY FOR GENERAL USE
# ===========================================================================
for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") in ["connector", "backshell"]:
        if instance.get("instance_name") not in ["X100"]:
            if instance.get("mpn") not in ["TXPA20"]:
                library_utils.pull(instance)

# ===========================================================================
#                  PROCESS HARNESS LAYOUT GRAPH
# ===========================================================================
formboard_utils.validate_nodes()

instances = fileio.read_tsv("instances list")
for instance in instances:
    if instance.get("item_type") == "cable":
        for instance2 in instances:
            if instance2.get("parent_instance") == instance.get("instance_name"):
                if instance2.get("item_type") == "conductor":
                    instances_list.modify(
                        instance.get("instance_name"),
                        {{
                            "node_at_end_a": instances_list.instance_in_connector_group_with_item_type(
                                instances_list.attribute_of(
                                    instance2.get("node_at_end_a"), "connector_group"
                                ),
                                "node",
                            ).get("instance_name"),
                            "node_at_end_b": instances_list.instance_in_connector_group_with_item_type(
                                instances_list.attribute_of(
                                    instance2.get("node_at_end_b"), "connector_group"
                                ),
                                "node",
                            ).get("instance_name"),
                        }},
                    )
                    break

for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") in ["conductor", "cable", "net-channel"]:
        formboard_utils.map_instance_to_segments(instance)

for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") in ["conductor", "cable"]:
        length = 0
        for instance2 in fileio.read_tsv("instances list"):
            if instance2.get("parent_instance") == instance.get("instance_name"):
                if instance2.get("length", "").strip():
                    length += int(instance2.get("length").strip())
        instances_list.modify(instance.get("instance_name"), {{"length": length}})

# ===========================================================================
#                   ASSIGN BOM LINE NUMBERS
# ===========================================================================
for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") in ["connector", "cable", "backshell"]:
        instances_list.modify(instance.get("instance_name"), {{"bom_line_number": True}})
instances_list.assign_bom_line_numbers()

# ===========================================================================
#       ASSIGN PRINT NAMES
# ===========================================================================
for x in range(2):
    for instance in fileio.read_tsv("instances list"):
        if instance.get("item_type") == "connector_cavity":
            instance_name = instance.get("instance_name", "")
            print_name = f"cavity {{instance_name.split(".")[-1] if "." in instance_name else instance_name}}"
            instances_list.modify(instance_name, {{"print_name": print_name}})

        elif instance.get("item_type") in ["conductor", "conductor-segment"]:
            instances_list.modify(instance.get("instance_name"), {{
                "print_name": f"'{{instance.get('cable_identifier')}}' of '{{instances_list.attribute_of(instance.get('cable_group'), 'print_name')}}'"
            }})

        elif instance.get("item_type") == "net-channel":
            print_name = f"'{{instance.get('this_channel_from_device_channel_id')}}' of '{{instance.get('this_channel_from_device_refdes')}}' to '{{instance.get('this_channel_to_device_channel_id')}}' of '{{instance.get('this_channel_to_device_refdes')}}'"
            instances_list.modify(instance.get("instance_name"), {{"print_name": print_name}})

        elif instance.get("item_type") == "net-channel-segment":
            print_name = f"'{{instances_list.attribute_of(instance.get('parent_instance'), 'this_channel_from_device_channel_id')}}' of '{{instances_list.attribute_of(instance.get('parent_instance'), 'this_channel_from_device_refdes')}}' to '{{instances_list.attribute_of(instance.get('parent_instance'), 'this_channel_to_device_channel_id')}}' of '{{instances_list.attribute_of(instance.get('parent_instance'), 'this_channel_to_device_refdes')}}'"
            instances_list.modify(instance.get("instance_name"), {{"print_name": print_name}})

        elif instance.get("item_type") == "connector":
            print_name = f"{{instance.get('connector_group')}}"
            instances_list.modify(instance.get("instance_name"), {{"print_name": print_name}})

        elif instance.get("item_type") == "cable-segment":
            print_name = f"{{instance.get('cable_group')}}"
            instances_list.modify(instance.get("instance_name"), {{"print_name": print_name}})

        elif instance.get("item_type") == "contact":
            print_name = instance.get("mpn")
            instances_list.modify(
                instance.get("instance_name"),
                {{"print_name": print_name}}
            )

        else:
            instances_list.modify(instance.get("instance_name"), {{"print_name": instance.get("instance_name")}})

# ===========================================================================
#                  ADD BUILD NOTES
# ===========================================================================
for rev_row in fileio.read_tsv("revision history"):
    if rev_row.get("rev") == state.rev:
        note_utils.make_rev_history_notes(rev_row)

for instance in fileio.read_tsv("instances list"):
    for note in note_utils.get_lib_build_notes(instance):
        note_utils.new_note(
            "build_note",
            note,
            affectedinstances=[instance.get("instance_name")]
        )

note_utils.assign_buildnote_numbers()

# example: add notes to describe actions
# note_utils.new_note(
#     "build_note",
#     "do this",
#     affectedinstances=["X1.B.conn"]
# )
# note_utils.new_note(
#     "build_note",
#     "do that"
# )

# example: combine buildnotes if their texts are similar
#note_utils.combine_notes("Torque backshell to connector at 40 in-lbs","Torque backshell to connector at about 40 in-lbs")


# ===========================================================================
#                  PUT TOGETHER FORMBOARD SVG INSTANCE CONTENT
# ===========================================================================
instances = fileio.read_tsv("instances list")
note_instances = []
for instance in instances:
    if instance.get("item_type") == "note":
        note_instances.append(note_utils.parse_note_instance(instance))

formboard_overview_instances = []
formboard_detail_instances = []
for instance in instances:
    if instance.get("item_type") not in ["connector", "backshell", "segment", "node", "origin"]:
        continue

    formboard_overview_instances.append(instance)
    formboard_detail_instances.append(instance)

    detail_flag_note_counter = 1
    overview_flag_note_counter = 1

    if instance.get("item_type") in ["connector", "backshell"]:
        formboard_detail_instances.append(note_utils.make_bom_flagnote(instance, f"flagnote-{{detail_flag_note_counter}}"))
        detail_flag_note_counter += 1

        formboard_detail_instances.append(note_utils.make_part_name_flagnote(instance, f"flagnote-{{detail_flag_note_counter}}"))
        detail_flag_note_counter += 1

    if instance.get("item_type") == "connector":
        formboard_overview_instances.append(note_utils.make_part_name_flagnote(instance, f"flagnote-{{overview_flag_note_counter}}"))
        overview_flag_note_counter += 1

    for note_instance in note_instances:
        if note_instance.get("note_type") == "build_note":
            if instance.get("instance_name") in note_instance.get("note_affected_instances"):
                formboard_detail_instances.append(
                    note_utils.make_buildnote_flagnote(note_instance, instance, f"flagnote-{{detail_flag_note_counter}}")
                )
                detail_flag_note_counter += 1

        if note_instance.get("note_type") == "rev_change_callout":
            if instance.get("instance_name") in note_instance.get("note_affected_instances"):
                formboard_detail_instances.append(
                    note_utils.make_rev_change_flagnote(note_instance, instance, f"flagnote-{{detail_flag_note_counter}}")
                )
                detail_flag_note_counter += 1

# ===========================================================================
#                  BUILD HARNESS OUTPUTS
# ===========================================================================
instances = fileio.read_tsv("instances list")
scales = {{"A": 0.25, "B": 0.3, "C": 1}}

feature_tree_utils.run_macro(
    "bom_exporter_bottom_up",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="bom-1",
)
feature_tree_utils.run_macro(
    "standard_harnice_formboard",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="formboard-overview",
    scale=scales.get("A"),
    input_instances=formboard_overview_instances,
)
feature_tree_utils.run_macro(
    "standard_harnice_formboard",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="formboard-detail",
    scale=scales.get("C"),
    input_instances=formboard_detail_instances,
)
feature_tree_utils.run_macro(
    "segment_visualizer",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="cable_layout-1",
    scale=scales.get("A"),
    item_type="cable-segment",
    segment_spacing_inches=0.1,
)
feature_tree_utils.run_macro(
    "segment_visualizer",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="conductor-layout-1",
    scale=scales.get("A"),
    item_type="conductor-segment",
    segment_spacing_inches=0.1,
)
feature_tree_utils.run_macro(
    "segment_visualizer",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="channel-layout-1",
    scale=scales.get("B"),
    item_type="net-channel-segment",
    segment_spacing_inches=0.1,
)
feature_tree_utils.run_macro(
    "circuit_visualizer",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="circuitviz-1",
    input_circuits=instances,
)
feature_tree_utils.run_macro(
    "revision_history_table",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="revhistory-1",
)

build_notes_list_instances = []
for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") == "note" and instance.get("note_type") == "build_note":
        build_notes_list_instances.append(instance)

feature_tree_utils.run_macro(
    "build_notes_table",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="build_notes_table-1",
    input_instances=build_notes_list_instances,
)
feature_tree_utils.run_macro(
    "pdf_generator",
    "harness_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="pdf_drawing-1",
    scales=scales,
)

{push_block}

# for convenience, move any pdf to the base directory of the harness
feature_tree_utils.copy_pdfs_to_cwd()
"""


# ======================================================================
# BLOCKS FOR BUILDING THE HARNESS FROM A SYSTEM
# ======================================================================


def default_build_macro_block(system_pn, system_rev, target_net):
    return f"""
# ===========================================================================
#                   build_macro SCRIPTING
# ===========================================================================
system_pn = "{system_pn}" # enter your system part number
system_rev = "{system_rev}" # enter your system revision
system_base_directory = fileio.get_path_to_project(system_pn) # add the path to project_locations.csv in the root of harnice
system_target_net = "{target_net}" # enter the net you're building from

feature_tree_utils.run_macro(
    "import_harness_from_harnice_system",
    "harness_builder",
    "https://github.com/harnice/harnice",
    "harness-from-system-1",
    system_pn=f"{{system_pn}}",
    system_rev=f"{{system_rev}}",
    path_to_system_rev=os.path.join(
        system_base_directory,
        f"{{system_pn}}-{{system_rev}}",
    ),
    target_net=system_target_net,
    manifest_nets=[system_target_net],
)

rev_history.overwrite(
    {{
        "desc": f"HARNESS '{{system_target_net}}' FROM SYSTEM '{{system_pn}}-{{system_rev}}'",
    }}
)
"""


def default_push_block():
    return """
# ensure the system that this harness was built from contains the complete updated instances list
post_harness_instances_list.push(
    system_base_directory,
    (system_pn, system_rev),
)
"""



import os
import csv
from collections import deque
from harnice import fileio
from harnice.lists import instances_list
from harnice.utils import library_utils


def mpn_of_device_refdes(refdes):
    """
    Looks up manufacturer part number information for a device reference designator.

    Searches the BOM (Bill of Materials) for a device with the given reference designator
    and returns its manufacturer, part number, and revision.

    **Args:**
    - `refdes` (str): Device reference designator to look up (e.g., `"J1"`, `"X1"`).

    **Returns:**
    - `tuple`: A tuple of `(MFG, MPN, rev)` if found, or `(None, None, None)` if not found.
    """
    for row in fileio.read_tsv("bom"):
        if row.get("device_refdes") == refdes:
            return row.get("MFG"), row.get("MPN"), row.get("rev")
    return None, None, None


def connector_of_channel(key):
    """
    Finds the connector name associated with a device channel.

    Given a device reference designator and channel ID tuple, looks up the corresponding
    connector name from the device's signals list.

    **Args:**
    - `key` (tuple): A tuple of `(device_refdes, channel_id)` identifying the channel.

    **Returns:**
    - `str`: The connector name associated with the channel.

    **Raises:**
    - `ValueError`: If the connector is not found for the given channel.
    """
    refdes, channel_id = key

    device_signals_list_path = os.path.join(
        fileio.dirpath("instance_data"),
        "device",
        refdes,
        f"{refdes}-signals_list.tsv",
    )
    for row in fileio.read_tsv(device_signals_list_path):
        if row.get("channel_id", "").strip() == channel_id.strip():
            return row.get("connector_name", "").strip()

    raise ValueError(f"Connector not found for channel {key}")


def find_connector_with_no_circuit(connector_list, circuits_list):
    """
    Validates that all connectors have associated circuits.

    Checks each connector in the connector list to ensure it has at least one
    corresponding circuit in the circuits list. Skips connectors with `"unconnected"`
    in their net name. Raises an error if any connector lacks a circuit.

    **Args:**
    - `connector_list` (list): List of connector dictionaries from the system connector list.
    - `circuits_list` (list): List of circuit dictionaries from the circuits list.

    **Raises:**
    - `ValueError`: If a connector is found that has no associated circuits. The error
        message suggests checking the channel map and channel compatibility.
    """
    for connector in connector_list:
        device_refdes = connector.get("device_refdes", "").strip()
        connector_name = connector.get("connector", "").strip()

        # skip if either key is missing
        if not device_refdes or not connector_name:
            continue

        # skip device if net name contains "unconnected"
        if "unconnected" in connector.get("net", "").strip():
            continue

        found_match = False
        for circuit in circuits_list:
            from_device_refdes = circuit.get("net_from_refdes", "").strip()
            from_connector_name = circuit.get("net_from_connector_name", "").strip()
            to_device_refdes = circuit.get("net_to_refdes", "").strip()
            to_connector_name = circuit.get("net_to_connector_name", "").strip()

            if (
                from_device_refdes == device_refdes
                and from_connector_name == connector_name
            ) or (
                to_device_refdes == device_refdes
                and to_connector_name == connector_name
            ):
                found_match = True
                break

        if not found_match:
            raise ValueError(
                f"Connector '{connector_name}' of device '{device_refdes}' does not contain any circuits. Check if it was mapped in the channel map. If n, check channel compatibility."
            )


def make_instances_for_connectors_cavities_nodes_channels_circuits():
    """
    Creates instances for all system components based on circuits.

    This function processes the circuits list and creates instances in the instances list
    for all connectors, connector cavities, nodes, channels, and circuits in the system.
    For each circuit, it creates:

    - Connector nodes (at both ends)
    - Connector instances (at both ends, with MPN lookup from system connector list)
    - Connector cavity instances (at both ends)
    - Circuit instance
    - Channel instance
    - Net-channel instances for nets in the channel chain (only for nets matching the circuit's net)

    After processing all circuits, the function updates connector instances with mating device
    information from the system connector list.

    The function reads from the circuits list, system connector list, and channel map
    to build the complete instance hierarchy.
    """
    connectors_list = fileio.read_tsv("system connector list")
    channel_map = fileio.read_tsv("channel map")

    for circuit in fileio.read_tsv("circuits list"):
        from_connector_key = (
            f"{circuit.get('net_from_refdes')}.{circuit.get('net_from_connector_name')}"
        )
        from_cavity = f"{from_connector_key}.{circuit.get('net_from_cavity')}"

        # Look up connector MPN from system connector list for the from connector
        from_connector_mpn = ""
        for connector in connectors_list:
            if connector.get("device_refdes") == circuit.get(
                "net_from_refdes"
            ) and connector.get("connector") == circuit.get("net_from_connector_name"):
                from_connector_mpn = connector.get("connector_mpn")
                break

        # from connector node
        instances_list.new_instance(
            f"{from_connector_key}.node",
            {
                "net": circuit.get("net"),
                "item_type": "node",
                "location_type": "node",
                "connector_group": from_connector_key,
            },
            ignore_duplicates=True,
        )

        # from connector
        instances_list.new_instance(
            f"{from_connector_key}.conn",
            {
                "net": circuit.get("net"),
                "item_type": "connector",
                "location_type": "node",
                "connector_group": from_connector_key,
                "parent_csys_instance_name": f"{from_connector_key}.node",
                "parent_csys_outputcsys_name": "origin",
                "this_instance_mating_device_refdes": circuit.get("net_from_refdes"),
                "this_instance_mating_device_connector": circuit.get(
                    "net_from_connector_name"
                ),
                "this_instance_mating_device_connector_mpn": from_connector_mpn,
            },
            ignore_duplicates=True,
        )

        # from connector cavity
        instances_list.new_instance(
            from_cavity,
            {
                "net": circuit.get("net"),
                "item_type": "connector_cavity",
                "parent_instance": f"{from_connector_key}.conn",  # from connector instance
                "location_type": "node",
                "connector_group": from_connector_key,
                "circuit_id": circuit.get("circuit_id"),
                "circuit_port_number": 0,
            },
            ignore_duplicates=True,
        )

        to_connector_key = (
            f"{circuit.get('net_to_refdes')}.{circuit.get('net_to_connector_name')}"
        )
        to_cavity = f"{to_connector_key}.{circuit.get('net_to_cavity')}"

        # Look up connector MPN from system connector list for the to connector
        to_connector_mpn = ""
        for connector in connectors_list:
            if connector.get("device_refdes") == circuit.get(
                "net_to_refdes"
            ) and connector.get("connector") == circuit.get("net_to_connector_name"):
                to_connector_mpn = connector.get("connector_mpn")
                break

        # to connector node
        instances_list.new_instance(
            f"{to_connector_key}.node",
            {
                "net": circuit.get("net"),
                "item_type": "node",
                "location_type": "node",
                "connector_group": to_connector_key,
            },
            ignore_duplicates=True,
        )

        # to connector
        instances_list.new_instance(
            f"{to_connector_key}.conn",
            {
                "net": circuit.get("net"),
                "item_type": "connector",
                "location_type": "node",
                "connector_group": to_connector_key,
                "parent_csys_instance_name": f"{to_connector_key}.node",
                "parent_csys_outputcsys_name": "origin",
                "this_instance_mating_device_refdes": circuit.get("net_to_refdes"),
                "this_instance_mating_device_connector": circuit.get(
                    "net_to_connector_name"
                ),
                "this_instance_mating_device_connector_mpn": to_connector_mpn,
            },
            ignore_duplicates=True,
        )

        # to connector cavity
        instances_list.new_instance(
            to_cavity,
            {
                "net": circuit.get("net"),
                "item_type": "connector_cavity",
                "parent_instance": f"{to_connector_key}.conn",  # to connector instance
                "location_type": "node",
                "connector_group": to_connector_key,
                "circuit_id": circuit.get("circuit_id"),
                "circuit_port_number": 1,
            },
            ignore_duplicates=True,
        )

        # add circuit
        instances_list.new_instance(
            f"circuit-{circuit.get('circuit_id')}",
            {
                "net": circuit.get("net"),
                "item_type": "circuit",
                "print_name": f"{circuit.get('signal')} of {circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')} <-> {circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}",
                "channel_group": f"channel-{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')}-{circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}",
                "circuit_id": circuit.get("circuit_id"),
                "node_at_end_a": from_cavity,
                "node_at_end_b": to_cavity,
                "this_net_from_device_refdes": circuit.get("net_from_refdes"),
                "this_net_from_device_channel_id": circuit.get("net_from_channel_id"),
                "this_net_from_device_connector_name": circuit.get(
                    "net_from_connector_name"
                ),
                "this_net_to_device_refdes": circuit.get("net_to_refdes"),
                "this_net_to_device_channel_id": circuit.get("net_to_channel_id"),
                "this_net_to_device_connector_name": circuit.get(
                    "net_to_connector_name"
                ),
                "this_channel_from_device_refdes": circuit.get(
                    "from_side_device_refdes"
                ),
                "this_channel_from_device_channel_id": circuit.get(
                    "from_side_device_chname"
                ),
                "this_channel_to_device_refdes": circuit.get("to_side_device_refdes"),
                "this_channel_to_device_channel_id": circuit.get(
                    "to_side_device_chname"
                ),
                "this_channel_from_channel_type": circuit.get("from_channel_type"),
                "this_channel_to_channel_type": circuit.get("to_channel_type"),
                "signal_of_channel_type": circuit.get("signal"),
            },
            ignore_duplicates=True,
        )

        # --- add channel
        instances_list.new_instance(
            f"channel-{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')}-"
            f"{circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}",
            {
                "item_type": "channel",
                "print_name": f"{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')} <-> {circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}",
                "channel_group": (
                    f"channel-{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')}-"
                    f"{circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}"
                ),
                "location_type": "segment",
                "this_channel_from_device_refdes": circuit.get(
                    "from_side_device_refdes"
                ),
                "this_channel_from_device_channel_id": circuit.get(
                    "from_side_device_chname"
                ),
                "this_channel_to_device_refdes": circuit.get("to_side_device_refdes"),
                "this_channel_to_device_channel_id": circuit.get(
                    "to_side_device_chname"
                ),
                "this_channel_from_channel_type": circuit.get("from_channel_type"),
                "this_channel_to_channel_type": circuit.get("to_channel_type"),
            },
            ignore_duplicates=True,
        )

        # Find the chain of nets for this channel from the channel map
        # The chain_of_nets field contains semicolon-separated net names
        chain_of_nets = []
        for channel in channel_map:
            if (
                channel.get("from_device_refdes")
                == circuit.get("from_side_device_refdes")
                and channel.get("from_device_channel_id")
                == circuit.get("from_side_device_chname")
                and channel.get("to_device_refdes")
                == circuit.get("to_side_device_refdes")
                and channel.get("to_device_channel_id")
                == circuit.get("to_side_device_chname")
            ):
                chain_of_nets = (channel.get("chain_of_nets") or "").split(";")
                break

        # Create net-channel instances for each net in the chain
        for net in chain_of_nets:
            if circuit.get("net") == net:
                instances_list.new_instance(
                    f"{net}:channel-{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')}-"
                    f"{circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}",
                    {
                        "net": net,
                        "item_type": "net-channel",
                        "print_name": (
                            f"{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')} <-> {circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}"
                        ),
                        "channel_group": (
                            f"channel-{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')}-{circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}"
                        ),
                        "location_type": "segment",
                        "parent_instance": f"channel-{circuit.get('from_side_device_refdes')}.{circuit.get('from_side_device_chname')}-{circuit.get('to_side_device_refdes')}.{circuit.get('to_side_device_chname')}",
                        "node_at_end_a": from_cavity,
                        "node_at_end_b": to_cavity,
                        "this_net_from_device_refdes": circuit.get("net_from_refdes"),
                        "this_net_from_device_channel_id": circuit.get(
                            "net_from_channel_id"
                        ),
                        "this_net_from_device_connector_name": circuit.get(
                            "net_from_connector_name"
                        ),
                        "this_net_to_device_refdes": circuit.get("net_to_refdes"),
                        "this_net_to_device_channel_id": circuit.get(
                            "net_to_channel_id"
                        ),
                        "this_net_to_device_connector_name": circuit.get(
                            "net_to_connector_name"
                        ),
                        "this_channel_from_device_refdes": circuit.get(
                            "from_side_device_refdes"
                        ),
                        "this_channel_from_device_channel_id": circuit.get(
                            "from_side_device_chname"
                        ),
                        "this_channel_to_device_refdes": circuit.get(
                            "to_side_device_refdes"
                        ),
                        "this_channel_to_device_channel_id": circuit.get(
                            "to_side_device_chname"
                        ),
                        "this_channel_from_channel_type": circuit.get(
                            "from_channel_type"
                        ),
                        "this_channel_to_channel_type": circuit.get("to_channel_type"),
                    },
                    ignore_duplicates=True,
                )

    # Post-process: Update all connector instances with mating device information
    # This ensures connector instances have the correct mating device refdes, connector name, and MPN
    for connector in fileio.read_tsv("system connector list"):
        try:
            instances_list.modify(
                f"{connector.get('device_refdes')}.{connector.get('connector')}.conn",
                {
                    "this_instance_mating_device_refdes": connector.get(
                        "device_refdes"
                    ),
                    "this_instance_mating_device_connector": connector.get("connector"),
                    "this_instance_mating_device_connector_mpn": connector.get(
                        "connector_mpn"
                    ),
                },
            )
        except ValueError:
            # Skip if connector instance doesn't exist (may not have been created if no circuits)
            pass


def add_chains_to_channel_map():
    """
    For each (from_device/channel) -> (to_device/channel) in the channel map,
    find:
      - 'disconnect_refdes_requirement' (like "X1(A,B);X2(B,A)")
      - 'chain_of_nets' (like "WH-1;WH-2;WH-3" or a single net if no disconnects)
      - 'chain_of_connectors' (like "WH-1.net.connector_name;WH-1.net.connector_name;WH-2.net.connector_name;WH-2.net.connector_name")
      - Format: net_name.connector_name where connector_name is net.connector_name (e.g., "/MIC_CABLE_2.MIC3out1" where "/MIC_CABLE_2" is the net)
      - Dots separate net_name from connector_name, semicolons separate connectors
    """

    channel_map = fileio.read_tsv("channel map")
    by_device, by_net, net_of, is_disconnect = {}, {}, {}, set()

    for row in fileio.read_tsv("system connector list"):
        dev = (row.get("device_refdes") or "").strip()
        con = (row.get("connector") or "").strip()
        net = (row.get("net") or "").strip()
        if not dev or not con:
            continue

        by_device.setdefault(dev, []).append(con)
        if net:
            by_net.setdefault(net, []).append((dev, con))
            net_of[(dev, con)] = net
        if (row.get("disconnect") or "").strip().upper() == "TRUE":
            is_disconnect.add(dev)

    def _shortest_disconnect_chain(from_cn_key, to_cn_key):
        start, goal = from_cn_key, to_cn_key
        q, seen, prev = deque([start]), {start}, {}

        while q:
            cur = q.popleft()
            if cur == goal:
                break
            net = net_of.get(cur)
            if net:
                for nxt in by_net.get(net, []):
                    if nxt not in seen:
                        seen.add(nxt)
                        prev[nxt] = cur
                        q.append(nxt)
            dev, _ = cur
            for other_con in by_device.get(dev, []):
                nxt = (dev, other_con)
                if nxt not in seen:
                    seen.add(nxt)
                    prev[nxt] = cur
                    q.append(nxt)

        if start != goal and goal not in prev:
            return [], [], []

        # reconstruct path
        path = [goal]
        while path[-1] != start:
            path.append(prev[path[-1]])
        path.reverse()

        # -------------------------------------------------------------
        # build disconnect chain, net chain, and connector chain
        chain, net_chain, connector_chain = [], [], []

        for a, b in zip(path, path[1:]):
            net_a = net_of.get(a)

            # collect unique net chain
            if net_a and (not net_chain or net_chain[-1] != net_a):
                net_chain.append(net_a)

            # detect disconnect traversal
            if a[0] == b[0] and a[0] in is_disconnect:
                dev = a[0]
                port_a, port_b = a[1], b[1]
                chain.append(f"{dev}({port_a},{port_b})")

        # Build connector chain in format: net.connector_name;net.connector_name;...
        # Format: {net}.{connector_name} where net comes from net_of lookup
        # For each net in the chain, find its first and last connector in the path
        net_to_connectors = {}
        for i, (dev, con) in enumerate(path):
            net = net_of.get((dev, con))
            if net:
                if net not in net_to_connectors:
                    net_to_connectors[net] = []
                net_to_connectors[net].append((i, dev, con))

        # Build connector chain: for each net in net_chain, add net.connector_name pairs
        # Format: {net}.{connector_name} (e.g., "WH-1.MIC3out1" where "WH-1" is the net and "MIC3out1" is the connector)
        for net in net_chain:
            if net in net_to_connectors:
                connectors = net_to_connectors[net]
                if len(connectors) >= 1:
                    # First connector is the from connector
                    _, from_dev, from_con = connectors[0]
                    # Last connector is the to connector
                    _, to_dev, to_con = connectors[-1]
                    # Look up the net for each connector (net_of maps (device_refdes, connector) -> net)
                    from_net = net_of.get((from_dev, from_con), "")
                    to_net = net_of.get((to_dev, to_con), "")
                    # Add both connectors: from connector and to connector for this net
                    connector_chain.append(f"{from_net}.{from_con}")
                    connector_chain.append(f"{to_net}.{to_con}")

        return chain, net_chain, connector_chain

    # -------------------------------------------------------------
    # apply to each channel pair
    for row in channel_map:
        from_key = (row.get("from_device_refdes"), row.get("from_device_channel_id"))
        to_key = (row.get("to_device_refdes"), row.get("to_device_channel_id"))
        if not all(from_key) or not all(to_key):
            continue

        from_cn = (from_key[0], connector_of_channel(from_key))
        to_cn = (to_key[0], connector_of_channel(to_key))

        n_from = net_of.get(from_cn)
        n_to = net_of.get(to_cn)

        # Special case: from and to connectors are on the same net (direct connection, no disconnects)
        if n_from and n_to and n_from == n_to:
            row["disconnect_refdes_requirement"] = ""
            row["chain_of_nets"] = n_from
            # Format: net.net.connector_name;net.net.connector_name
            # Look up the net for each connector (should match n_from/n_to)
            from_net = net_of.get(from_cn, "")
            to_net = net_of.get(to_cn, "")
            row["chain_of_connectors"] = (
                f"{n_from}.{from_net}.{from_cn[1]};{n_to}.{to_net}.{to_cn[1]}"
            )
            continue

        chain, net_chain, connector_chain = _shortest_disconnect_chain(from_cn, to_cn)
        if chain or net_chain or connector_chain:
            row["disconnect_refdes_requirement"] = ";".join(chain)
            row["chain_of_nets"] = ";".join(net_chain)
            row["chain_of_connectors"] = ";".join(connector_chain)

    with open(fileio.path("channel map"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=channel_map[0].keys(), delimiter="\t")
        writer.writeheader()
        writer.writerows(channel_map)


def make_instances_from_bom():
    """
    Creates instances for all devices and disconnects from the BOM.

    Reads the Bill of Materials (BOM) and imports each device or disconnect into the
    instances list using `library_utils.pull()`. Each item is imported with its manufacturer,
    part number, revision, and library information.

    Items with the `"disconnect"` field set are imported as type `"disconnect"`,
    all others are imported as type `"device"`.
    """
    for device in fileio.read_tsv("bom"):
        if device.get("disconnect"):
            item_type = "disconnect"
        else:
            item_type = "device"

        library_utils.pull(
            {
                "instance_name": device.get("device_refdes"),
                "mfg": device.get("MFG"),
                "mpn": device.get("MPN"),
                "item_type": item_type,
                "lib_repo": device.get("lib_repo"),
                "lib_subpath": device.get("lib_subpath"),
                "lib_rev_used_here": device.get("rev"),
            }
        )



import runpy
import os
import csv
from harnice import fileio, state
from harnice.products import chtype
from harnice.lists import signals_list

default_desc = "DISCONNECT, FUNCTION, ATTRIBUTES, etc."

disconnect_feature_tree_utils_default = """
from harnice.lists import signals_list
from harnice.products import chtype

ch_type_ids = {
    "A": {
        "balanced audio mic level in": (1, "https://github.com/harnice/harnice"),
        "chassis": (5, "https://github.com/harnice/harnice")
    },
    "B": {
        "balanced audio mic level out": (2, "https://github.com/harnice/harnice"),
        "chassis": (5, "https://github.com/harnice/harnice")
    }
}

cn_mpns = {
    "A": "DB25F",
    "B": "DB25M"
}

cavity_number = {
    "ch0": {
        "pos": 24,
        "neg": 12,
        "chassis": 25
    },
    "ch1": {
        "pos": 10,
        "neg": 23,
        "chassis": 11
    },
    "ch2": {
        "pos": 21,
        "neg": 9,
        "chassis": 22
    },
    "ch3": {
        "pos": 7,
        "neg": 20,
        "chassis": 8
    },
    "ch4": {
        "pos": 18,
        "neg": 6,
        "chassis": 19
    },
    "ch5": {
        "pos": 4,
        "neg": 17,
        "chassis": 5
    },
    "ch6": {
        "pos": 15,
        "neg": 3,
        "chassis": 16
    },
    "ch7": {
        "pos": 1,
        "neg": 14,
        "chassis": 2
    },
}

signals_list.new()

for channel in range(8):
    channel_name = f"ch{channel}"

    for signal in chtype.signals(ch_type_ids["A"]["balanced audio mic level in"]):
        signals_list.append(
            channel_id=channel_name,
            signal=signal,

            A_cavity=cavity_number[channel_name][signal],
            A_connector_mpn=cn_mpns["A"],
            A_channel_type=ch_type_ids["A"]["balanced audio mic level in"],

            B_cavity=cavity_number[channel_name][signal],
            B_connector_mpn=cn_mpns["B"],
            B_channel_type=ch_type_ids["B"]["balanced audio mic level out"],
        )

    for signal in chtype.signals(ch_type_ids["A"]["chassis"]):
        signals_list.append(
            channel_id=f"{channel_name}-shield",
            signal=signal,

            A_cavity=cavity_number[channel_name][signal],
            A_connector_mpn=cn_mpns["A"],
            A_channel_type=ch_type_ids["A"]["chassis"],

            B_cavity=cavity_number[channel_name][signal],
            B_connector_mpn=cn_mpns["B"],
            B_channel_type=ch_type_ids["B"]["chassis"],
        )

"""


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-feature_tree.py": "feature tree",
        f"{state.partnumber('pn-rev')}-signals_list.tsv": "signals list",
        f"{state.partnumber('pn-rev')}-attributes.json": "attributes",
    }


def generate_structure():
    pass


def _validate_signals_list():
    print("--------------------------------")
    print("Validating signals list...")
    if not os.path.exists(fileio.path("signals list")):
        raise FileNotFoundError("Signals list was not generated.")

    with open(fileio.path("signals list"), "r", encoding="utf-8") as f:
        reader = csv.DictReader(f, delimiter="\t")
        headers = reader.fieldnames
        signals_list = list(reader)

    if not headers:
        raise ValueError("Signals list has no header row.")

    counter = 2
    for signal in signals_list:
        print("Looking at csv row:", counter)
        A_channel_type = chtype.parse(signal.get("A_channel_type"))
        B_channel_type = chtype.parse(signal.get("B_channel_type"))

        # make sure all the fields are there
        if signal.get("channel_id") in ["", None]:
            raise ValueError("A_channel_id is blank")
        if signal.get("signal") in ["", None]:
            raise ValueError("signal is blank")
        if signal.get("A_cavity") in ["", None]:
            raise ValueError("A_cavity is blank")
        if signal.get("B_cavity") in ["", None]:
            raise ValueError("B_cavity is blank")
        if signal.get("A_connector_mpn") in ["", None]:
            raise ValueError("A_connector_mpn is blank")
        if signal.get("A_channel_type") in ["", None]:
            raise ValueError("A_channel_type is blank")
        if signal.get("B_connector_mpn") in ["", None]:
            raise ValueError("B_connector_mpn is blank")
        if signal.get("B_channel_type") in ["", None]:
            raise ValueError("B_channel_type is blank")

        # make sure signal is a valid signal of its channel type
        if signal.get("signal") not in chtype.signals(A_channel_type):
            raise ValueError(
                f"Signal {signal.get('A_signal')} is not a valid signal of its channel type"
            )

        # make sure A and B sides are compatible
        if B_channel_type not in chtype.compatibles(A_channel_type):
            if A_channel_type not in chtype.compatibles(B_channel_type):
                raise ValueError("A and B channel types are not compatible")

        expected_signals = chtype.signals(A_channel_type)
        found_signals = set()

        # make sure all the signals of each channel type are present
        for expected_signal in expected_signals:
            for signal2 in signals_list:
                if (
                    signal2.get("channel_id") == signal.get("channel_id")
                    and signal2.get("signal") == expected_signal
                ):
                    found_signals.add(expected_signal)

        missing_signals = set(expected_signals) - found_signals
        if missing_signals:
            raise ValueError(
                f"Channel {signal.get('channel_id')} is missing signals: {', '.join(missing_signals)}"
            )

        counter += 1

    # make sure no duplicate A-side cavities are present
    seen_A = set()
    for signal in signals_list:
        A_cavity = signal.get("A_cavity")
        if A_cavity in seen_A:
            raise ValueError(f"Duplicate A_cavity found in disconnect: {A_cavity}")
        seen_A.add(A_cavity)

    # make sure no duplicate B-side cavities are present
    seen_B = set()
    for signal in signals_list:
        B_cavity = signal.get("B_cavity")
        if B_cavity in seen_B:
            raise ValueError(f"Duplicate B_cavity found in disconnect: {B_cavity}")
        seen_B.add(B_cavity)

    if counter == 2:
        raise ValueError(
            "No signals have been specified. Check your feature tree or add rows manually."
        )

    print(f"Signals list of {state.partnumber('pn')} is valid.\n")


def render():
    signals_list.set_list_type("disconnect")
    if not os.path.exists(fileio.path("signals list")):
        if not os.path.exists(fileio.path("feature tree")):
            with open(fileio.path("feature tree"), "w", encoding="utf-8") as f:
                f.write(disconnect_feature_tree_utils_default)

    if os.path.exists(fileio.path("feature tree")):
        runpy.run_path(fileio.path("feature tree"))
        print("Successfully rebuilt signals list per feature tree.")

    _validate_signals_list()



import os
import csv
import ast
import importlib
from harnice import fileio, state

# === Global Columns Definition ===
COLUMNS = [
    "product",  # the harnice product type (e.g. "harness", "connector", "device", "system", "macro", "flagnote", "tblock")
    "mfg",  # who manufactures this product (blank ok)
    "pn",  # name, part number, other identifier of this part. mfg+mpn combination must be unique within the library.
    "desc",  # a brief description of this product
    "rev",  # the revision of the part
    "status",  # "released", "obsolete", etc. Harnice will not render a revision if the status has text in this field as a form of protection.
    "releaseticket",  # many companies do this, but it's not required.
    "library_repo",  # auto-filled on render if the current working directory is discovered to be a library repository.
    "library_subpath",  # auto-filled on render if in a library repository, this is the chain of directories between the product type and the part number
    "datestarted",  # auto-filled to be the date when this part was first intialized
    "datemodified",  # updates to today's date upon rendering
    "datereleased",  # up to user to fill in as needed
    "git_hash_of_harnice_src",  # auto-filled, git hash of the harnice source code during the latest render
    "drawnby",  # auto-filled, the person who created the part
    "checkedby",  # the person who checked the part, blank ok
    "revisionupdates",  # a brief description of the changes made to this revision
    "affectedinstances",  # the instance names of the instances that were affected by this revision. can be referenced later by PDF builders and more.
]


def overwrite(content_dict):
    """
Overwrite a revision history entry.

**Arguments:**

 - `content_dict` (dict): The content to overwrite the revision history entry with.
    - This should be a dictionary with the keys and values to overwrite.
    - The keys should be the column names, and the values should be the new values.
    - Some keys are protected and cannot be overwritten:
        - `"product"`
        - `"mfg"`
        - `"pn"`
        - `"rev"`
        - `"releaseticket"`
        - `"library_repo"`
        - `"library_subpath"`
        - `"datestarted"`

The function will update the revision history file as referenced by the current product file structure.

**Returns:**

 - `None`

**Raises:**

 - `KeyError`: If a key is provided that is not in the COLUMNS list.
 - `KeyError`: If a protected key is provided.
 - `ValueError`: If the revision history file is not found.
 - `ValueError`: If the revision is not found in the revision history file.
 - `RuntimeError`: If `state.rev` is not set.
    """
    PROTECTED_KEYS = [
        "product",
        "mfg",
        "pn",
        "rev",
        "releaseticket",
        "library_repo",
        "library_subpath",
        "datestarted",
    ]
    # 1) Ensure no unknown keys
    for key in content_dict:
        if key not in COLUMNS:
            raise KeyError(
                f"Harnice does not allow writing unknown key '{key}'. "
                f"Valid columns: {', '.join(COLUMNS)}"
            )

    # 2) Ensure none of the protected keys are being modified
    for key in PROTECTED_KEYS:
        if key in content_dict:
            raise KeyError(
                f"Harnice does not allow overwriting '{key}' by script.\n"
                f"Please edit the revision history manually."
            )

    # 3) Load or create revision history
    path = fileio.path("revision history")
    if not os.path.exists(path):
        new()  # Creates a blank rev history with header

    rows = fileio.read_tsv("revision history")

    # 4) Determine which revision we are updating from state
    target_rev = str(state.rev).strip()
    if not target_rev:
        raise RuntimeError("state.rev is not set. Did verify_revision_structure() run?")

    # 5) Update matching row
    found = False
    for row in rows:
        if str(row.get("rev", "")).strip() == target_rev:
            found = True
            for key, value in content_dict.items():
                row[key] = value

    if not found:
        raise ValueError(f"No revision '{target_rev}' found in revision history.")

    # 6) Write updated TSV
    with open(path, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(rows)


def info(rev=None, path=None, field=None, all=False):
    """
    Get information about a revision history entry.

    **Arguments:**

    - `rev` (str): The revision to get information about.
    - `path` (str): The path to the revision history file.
        - If not provided, the function will use the default path: `"revision history"`.
    - `field` (str): The field to get information about.
        - If not provided, the function will return the entire row.
        - If provided, the function will return the value of the field.
    - `all` (bool): If `True`, return all rows.
        - If not provided, the function will return the first row.

    **Returns:**

    - `dict`: The row of the revision history entry (when `field` is not provided).
    - `list`: A list of all rows in the revision history file (when `all=True`).
    - `str`: The value of the field (when `field` is provided).

    **Raises:**

    - `FileNotFoundError`: If the revision history file is not found.
    - `ValueError`: If the revision is not found in the revision history file.
    """
    if path is None:
        path = fileio.path("revision history")

    if not os.path.exists(path):
        raise FileNotFoundError(f"Revision history file not found at {path}")

    if rev:
        rev = str(rev)
    else:
        rev = state.partnumber("R")

    with open(path, newline="", encoding="utf-8") as f:
        reader = csv.DictReader(f, delimiter="\t")
        rows = list(reader)

    if all:
        return rows

    for row in rows:
        if row.get("rev") == rev:
            # ------------------------------------------------------
            # Field requested
            # ------------------------------------------------------
            if field:
                val = row.get(field)

                if field == "affectedinstances":
                    if not val or val.strip() == "":
                        return []
                    try:
                        return ast.literal_eval(val)
                    except Exception:
                        # fallback: return empty list if malformed
                        return []

                # other fields unchanged
                return val

            # ------------------------------------------------------
            # Entire row requested: parse affectedinstances only
            # ------------------------------------------------------
            full_row = dict(row)
            ai = row.get("affectedinstances")
            if ai and ai.strip() != "":
                try:
                    full_row["affectedinstances"] = ast.literal_eval(ai)
                except Exception:
                    full_row["affectedinstances"] = []
            else:
                full_row["affectedinstances"] = []

            return full_row

    raise ValueError(f"Revision {rev} not found in revision history at {path}")


def initial_release_exists():
    """
    Check if an initial release exists.

    **Arguments:**

    None

    **Returns:**

    - `bool`: `True` if a revision with the text `"INITIAL RELEASE"` in the `"revisionupdates"` field exists, `False` otherwise.
    """
    try:
        for row in fileio.read_tsv("revision history"):
            if str(row.get("revisionupdates", "")).strip() == "INITIAL RELEASE":
                return True
            else:
                return False
    except NameError:
        return False


def initial_release_desc():
    """
    Get the description of the initial release.

    **Arguments:**

    None

    **Returns:**

    - `str`: The description of the revision which has `revisionupdates == 'INITIAL RELEASE'`.
    """
    for row in fileio.read_tsv("revision history"):
        if row.get("revisionupdates") == "INITIAL RELEASE":
            return row.get("desc")


def update_datemodified():
    """
    Update the `datemodified` field of the current revision with today's date.

    **Arguments:**

    None

    **Returns:**

    - `None`

    **Raises:**

    - `ValueError`: If the revision history file is not found.
    - `ValueError`: If the revision is not found in the revision history file.
    """
    target_rev = state.partnumber("R")

    # Read all rows
    with open(fileio.path("revision history"), newline="", encoding="utf-8") as f_in:
        reader = csv.DictReader(f_in, delimiter="\t")
        rows = list(reader)

    # Modify matching row(s)
    for row in rows:
        if row.get("rev", "").strip() == target_rev:
            row["datemodified"] = fileio.today()
            row["drawnby"] = fileio.drawnby()["name"]
            row["git_hash_of_harnice_src"] = fileio.get_git_hash_of_harnice_src()

    # Write back
    with open(
        fileio.path("revision history"), "w", newline="", encoding="utf-8"
    ) as f_out:
        writer = csv.DictWriter(f_out, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(rows)


def new(ignore_product=False, path=None):
    """
    Create a new revision history file.

    **Arguments:**

    - `ignore_product` (bool):
        - If `True`, the function will raise an error if `state.product` is not set first.
        - If `False`, the function will prompt the user to select a product type.

    **Returns:**

    - `None`

    **Raises:**

    - `ValueError`: If attempting to create a new revision history file without a product type when `ignore_product=True`.
    - `ValueError`: If attempting to overwrite an existing revision history file.
    """
    columns = COLUMNS

    if path is None:
        path = fileio.path("revision history")

    if not ignore_product:
        from harnice.cli import select_product_type

        state.set_product(select_product_type())

    if ignore_product and not state.product:
        raise ValueError(
            "You tried to create a new revision history file without a product type. This is not allowed."
        )

    if not os.path.exists(path):
        with open(path, "w", newline="", encoding="utf-8") as f:
            writer = csv.DictWriter(f, fieldnames=columns, delimiter="\t")
            writer.writeheader()

    else:
        raise ValueError(
            "You tried to overwrite a revision history file- this is not allowed."
        )


def append(next_rev=None):
    """
    Append a new revision history entry to the current revision history file.

    If the revision history file does not exist, the function will create it.
    If the revision history file exists, the function will append a new entry to the file.

    It will prompt the user for the following fields:

    - `product`: The product type of the part.
    - `desc`: The description of the part.
    - `revisionupdates`: What is the purpose of this revision?

    If the previous revision has a blank status, the function will prompt the user to obsolete it with a message.

    **Arguments:**

    - `next_rev` The next revision number to append.

    **Returns:**

    - `None`
    """
    from harnice import cli

    if not os.path.exists(fileio.path("revision history")):
        new()
    rows = fileio.read_tsv("revision history")
    if rows:
        for row in reversed(rows):
            candidate = (row.get("product") or "").strip()
            if candidate:
                state.set_product(candidate)
                break
    if not state.product:
        state.set_product(cli.select_product_type())

    default_desc = ""
    if state.product:
        try:
            product_module = importlib.import_module(
                f"harnice.products.{state.product}"
            )
        except ModuleNotFoundError:
            product_module = None
        else:
            default_desc = getattr(product_module, "default_desc", "") or ""

    desc = ""
    if next_rev != 1:
        # find the highest revision in the table
        try:
            highest_existing_rev = max(
                int(row.get("rev", 0)) for row in rows if row.get("rev")
            )
        except ValueError:
            highest_existing_rev = None

        if next_rev != highest_existing_rev:
            for row in rows:
                if int(row.get("rev", 0)) == highest_existing_rev:
                    desc = row.get("desc")
                    if row.get("status") in [None, ""]:
                        print(
                            f"Your existing highest revision ({highest_existing_rev}) has no status. Do you want to obsolete it?"
                        )
                        obsolete_message = cli.prompt(
                            "Type your message here, leave blank for 'OBSOLETE' message, or type 'n' to keep it blank.",
                            default="OBSOLETE",
                        )
                        if obsolete_message == "n":
                            obsolete_message = ""
                        row["status"] = obsolete_message  # ← modified here
                    break

    if desc in [None, ""]:
        desc = cli.prompt(
            f"Enter a description of this {state.product}",
            default=default_desc,
        )

    revisionupdates = "INITIAL RELEASE"
    if initial_release_exists():
        revisionupdates = ""
    revisionupdates = cli.prompt(
        "Enter a description for this revision", default=revisionupdates
    )
    while not revisionupdates or not revisionupdates.strip():
        print("Revision updates can't be blank!")
        revisionupdates = cli.prompt(
            "Enter a description for this revision", default=None
        )

    # add lib_repo if filepath is found in library locations
    library_repo = ""
    library_subpath = ""
    cwd = str(os.getcwd()).lower().strip("~")

    for row in fileio.read_tsv("library locations", delimiter=","):
        lib_local_path = str(row.get("local_path", "")).lower().strip("~")
        if lib_local_path in cwd:
            library_repo = row.get("repo_url")

            # keep only the portion AFTER local_path
            idx = cwd.find(lib_local_path)
            remainder = cwd[idx + len(lib_local_path) :].lstrip("/")
            parts = remainder.split("/")

            # find the part number in the path
            pn = str(state.partnumber("pn")).lower()
            if pn in parts:
                pn_index = parts.index(pn)
                core_parts = parts[:pn_index]  # everything before pn
            else:
                core_parts = parts

            # build library_subpath and product
            if core_parts:
                library_subpath = (
                    "/".join(core_parts[1:]) + "/" if len(core_parts) > 1 else ""
                )  # strip out the first element (product type)
            else:
                library_subpath = ""

            break

    ####

    rows.append(
        {
            "product": state.product,
            "pn": state.pn,
            "rev": next_rev,
            "desc": desc,
            "status": "",
            "library_repo": library_repo,
            "library_subpath": library_subpath,
            "datestarted": fileio.today(),
            "datemodified": fileio.today(),
            "revisionupdates": revisionupdates,
        }
    )

    columns = COLUMNS
    with open(fileio.path("revision history"), "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=columns, delimiter="\t")
        writer.writeheader()
        writer.writerows(rows)


def part_family_append(content_dict, rev_history_path):
    """
    Append a new revision history entry to the part family revision history file.

    Intended to be called by part family scripts only.

    The function will automatically update the following fields in the content dictionary:

    - `datemodified`: Set to today's date
    - `drawnby`: Set to the current user's name
    - `git_hash_of_harnice_src`: Set to the current git hash of the harnice source code

    If the revision history file does not exist, the function will create it.
    If an entry with the same revision number already exists, the function will update that entry.
    Otherwise, the function will append a new entry to the file.

    **Arguments:**

    - `content_dict` (dict): The content to append to the part family revision history file.
        - Should contain keys matching the `COLUMNS` list.
        - The `rev` key is used to determine if an entry already exists.
    - `rev_history_path` (str): The path to the part family revision history file.

    **Returns:**

    - `None`

    **Raises:**

    - `ValueError`: If the content dictionary contains invalid keys or missing required fields.
    """
    actual_content_dict = content_dict

    actual_content_dict["datemodified"] = fileio.today()
    actual_content_dict["drawnby"] = fileio.drawnby()["name"]
    actual_content_dict["git_hash_of_harnice_src"] = (
        fileio.get_git_hash_of_harnice_src()
    )

    rev = content_dict.get("rev")

    if os.path.exists(rev_history_path):
        rows = fileio.read_tsv(rev_history_path)
    else:
        rows = []

    found = False
    for i, row in enumerate(rows):
        if row.get("rev") == actual_content_dict.get("rev"):
            rows[i] = actual_content_dict
            found = True
            break

    if not found:
        rows.append(actual_content_dict)

    if not os.path.exists(rev_history_path):
        new(path=rev_history_path, ignore_product=True)
    with open(rev_history_path, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=COLUMNS, delimiter="\t")
        writer.writeheader()
        writer.writerows(rows)



import os
import csv
import shutil
from harnice import fileio
from harnice.lists import instances_list


def rebuild():
    """
    Build the 'post harness instances list' by merging instance data from:
      - Each harness's instances list if the harness_pn is defined and file exists
      - Otherwise, fall back to the system-level instances list for matching nets

    Writes a clean TSV with INSTANCES_LIST_COLUMNS.
    """

    post_harness_instances = []

    # --- iterate through manifest rows ---
    for harness in fileio.read_tsv("harness manifest"):
        harness_pn = (harness.get("harness_pn") or "").strip()
        net = (harness.get("net") or "").strip()
        if not net:
            continue

        # Case 1: harness_pn missing -> import from system instances
        if not harness_pn:
            for system_instance in fileio.read_tsv("instances list"):
                if system_instance.get("net", "").strip() == net:
                    post_harness_instances.append(system_instance)
            continue

        # Case 2: harness_pn provided -> try to load harness instances list
        harness_instances_list_path = os.path.join(
            fileio.dirpath("harnesses"),
            f"{harness_pn}-instances_list.tsv",
        )

        if os.path.exists(harness_instances_list_path):
            post_harness_instances.extend(fileio.read_tsv(harness_instances_list_path))
        else:
            # Fallback to system-level instances for same net
            for system_instance in fileio.read_tsv("instances list"):
                if system_instance.get("net", "").strip() == net:
                    post_harness_instances.append(system_instance)

    # --- Determine fieldnames dynamically ---
    fieldnames = set()
    for instance in post_harness_instances:
        fieldnames.update(instance.keys())

    # --- Write TSV file ---
    with open(
        fileio.path("post harness instances list"), "w", newline="", encoding="utf-8"
    ) as f:
        writer = csv.DictWriter(f, fieldnames=instances_list.COLUMNS, delimiter="\t")
        writer.writeheader()
        for instance in post_harness_instances:
            writer.writerow({k: instance.get(k, "") for k in fieldnames})


def push(path_to_system_rev, system_pn_rev):
    path_to_harness_dir_of_system = os.path.join(
        path_to_system_rev, f"{system_pn_rev[0]}-{system_pn_rev[1]}", "harnesses"
    )
    shutil.copy(fileio.path("instances list"), path_to_harness_dir_of_system)



import runpy
import os
from harnice import fileio, state, cli
from harnice.lists import post_harness_instances_list, instances_list, library_history

default_desc = "SYSTEM, SCOPE, etc."

system_feature_tree_utils_default = """from harnice import fileio
from harnice.utils import system_utils, feature_tree_utils
from harnice.lists import instances_list, manifest, channel_map, circuits_list, disconnect_map

#===========================================================================
#                   KICAD PROCESSING
#===========================================================================
feature_tree_utils.run_macro("kicad_pro_to_bom", "system_builder", "https://github.com/harnice/harnice", artifact_id="bom-1")

#===========================================================================
#                   COLLECT AND PULL DEVICES FROM LIBRARY
#===========================================================================
system_utils.make_instances_from_bom()

#===========================================================================
#                   CHANNEL MAPPING
#===========================================================================
feature_tree_utils.run_macro(
    "kicad_pro_to_system_connector_list",
    "system_builder",
    "https://github.com/harnice/harnice",
    artifact_id="system-connector-list-1"
)
manifest.new()
channel_map.new()

#add manual channel map commands here. key=(from_device_refdes, from_device_channel_id)
#channel_map.map(("MIC3", "out1"), ("PREAMP1", "in2"))

#map channels to other compatible channels by sorting alphabetically then mapping compatibles
feature_tree_utils.run_macro(
    "basic_channel_mapper",
    "system_builder",
    "https://github.com/harnice/harnice",
    artifact_id="channel-mapper-1"
)

#if mapped channels must connect via disconnects, add the list of disconnects to the channel map
system_utils.add_chains_to_channel_map()

#map channels that must pass through disconnects to available channels inside disconnects
disconnect_map.new()

#add manual disconnect map commands here
#disconnect_map.already_assigned_disconnects_set_append(('X1', 'ch0'))

#map channels passing through disconnects to available channels inside disconnects
feature_tree_utils.run_macro(
    "disconnect_mapper",
    "system_builder",
    "https://github.com/harnice/harnice",
    artifact_id="disconnect-mapper-1"
)

# ensure channels that are required to be mapped through disconnects are in fact done so
disconnect_map.ensure_requirements_met()

#process channel and disconnect maps to make a list of every circuit in your system
circuits_list.new()

# ===========================================================================
#                   INSTANCES LIST
# ===========================================================================
system_utils.make_instances_for_connectors_cavities_nodes_channels_circuits()

# assign mating connectors
for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") == "connector":
        if instance.get("this_instance_mating_device_connector_mpn") == "XLR3M":
            instances_list.modify(
                instance.get("instance_name"),
                {
                    "mpn": "D38999_26ZA98PN",
                    "lib_repo": "https://github.com/harnice/harnice",
                },
            )
        elif instance.get("this_instance_mating_device_connector_mpn") == "XLR3F":
            instances_list.modify(
                instance.get("instance_name"),
                {
                    "mpn": "D38999_26ZB98PN",
                    "lib_repo": "https://github.com/harnice/harnice",
                },
            )
        elif instance.get("this_instance_mating_device_connector_mpn") == "DB25M":
            instances_list.modify(
                instance.get("instance_name"),
                {
                    "mpn": "D38999_26ZC35PN",
                    "lib_repo": "https://github.com/harnice/harnice",
                },
            )
        elif instance.get("this_instance_mating_device_connector_mpn") == "DB25F":
            instances_list.modify(
                instance.get("instance_name"),
                {
                    "mpn": "D38999_26ZE6PN",
                    "lib_repo": "https://github.com/harnice/harnice",
                },
            )

# ===========================================================================
#                   ASSIGN CONDUCTORS
# ===========================================================================

# add one conductor per circuit
for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") == "circuit":
        circuit_id = instance.get("circuit_id")
        conductor_name = f"conductor-{circuit_id}"
        instances_list.new_instance(
            conductor_name,
            {
                "net": instance.get("net"),
                "item_type": "conductor",
                "location_type": "segment",
                "channel_group": instance.get("channel_group"),
                "node_at_end_a": circuit_utils.instance_of_circuit_port_number(
                    circuit_id, 0 # assume the only existing ports at this point are cavities at 0 and 1
                ),
                "node_at_end_b": circuit_utils.instance_of_circuit_port_number(
                    circuit_id, 1
                ),
                "this_channel_from_channel_type": instance.get("this_channel_from_channel_type"),
                "this_channel_to_channel_type": instance.get("this_channel_to_channel_type"),
                "signal_of_channel_type": instance.get("signal_of_channel_type")
            },
        )
        circuit_utils.squeeze_instance_between_ports_in_circuit(
            conductor_name, instance.get("circuit_id"), 1
        )

# define the cable types we want to use here
audio_cable = {
    "lib_repo": "https://github.com/harnice/harnice",
    "mpn": "8762 0602000",
    "lib_subpath": "belden",
}

# assign conductors to cable-id
cable_id_counter = 1
instances = fileio.read_tsv("instances list")
for net in instances_list.list_of_uniques("net"):
    for chgroup in instances_list.list_of_uniques("channel_group"):
        for instance in instances:
            cable_name = f"cable-{cable_id_counter}"

            if instance.get("net") != net:
                continue
            if instance.get("channel_group") != chgroup:
                continue
            if instance.get("item_type") != "conductor":
                continue

            if chtype.parse(instance.get("this_channel_from_channel_type")) in chtype.is_or_is_compatible_with((1, 'https://github.com/harnice/harnice')):
                if instance.get("signal_of_channel_type") in ["pos"]:
                    circuit_utils.assign_cable_conductor(
                        cable_name,
                        ("pair_1", "white"),
                        instance.get("instance_name"),
                        audio_cable,
                        instance.get("net")
                    )
                if instance.get("signal_of_channel_type") in ["neg"]:
                    circuit_utils.assign_cable_conductor(
                        cable_name,
                        ("pair_1", "black"),
                        instance.get("instance_name"),
                        audio_cable,
                        instance.get("net")
                    )
        cable_id_counter += 1 # in this system, each channel gets its own cable.

#===========================================================================
#                   SYSTEM DESIGN CHECKS
#===========================================================================
connector_list = fileio.read_tsv("system connector list")
circuits_list = fileio.read_tsv("circuits list")

#check for circuits with no connectors
system_utils.find_connector_with_no_circuit(connector_list, circuits_list)


# ===========================================================================
#                   ADD STYLING AND PRINT NAMES TO CHANNELS
# ===========================================================================
# assign styling to channels
audio_channel_style = {
    "base_color": "#D59A10",
}
shield_channel_style = {
    "base_color": "#4039A1",
}

for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") in ["channel", "net-channel"]:
        if instance.get("this_channel_from_channel_type") in ["(1, 'https://github.com/harnice/harnice')", "(2, 'https://github.com/harnice/harnice')"]:
            instances_list.modify(
                instance.get("instance_name"),
                {
                    "appearance": audio_channel_style
                }
            )
        if instance.get("this_channel_from_channel_type") == "(5, 'https://github.com/harnice/harnice')":
            instances_list.modify(
                instance.get("instance_name"),
                {
                    "appearance": shield_channel_style
                }
            )

for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") == "net-channel":
        instances_list.modify(instance.get("instance_name"), {
            "print_name_at_end_a": instance.get("this_net_from_device_channel_id"),
            "print_name_at_end_b": instance.get("this_net_to_device_channel_id")
        })

# ===========================================================================
#                   SYSTEM ARTIFACT GENERATORS
# ===========================================================================

# prepare the channel map block diagram overlay
chmap_instances = []
for instance in fileio.read_tsv("instances list"):
    if instance.get("item_type") == "net-channel":
        chmap_instances.append(instance)
feature_tree_utils.run_macro(
    "kicad_sch_net_overlay",
    "system_artifacts",
    "https://github.com/harnice/harnice",
    artifact_id="blockdiagram-chmap-1",
    item_type="net-channel",
    instances=chmap_instances
)

# for convenience, move any pdf to the base directory of the harness
feature_tree_utils.copy_pdfs_to_cwd()
"""


def file_structure():
    return {
        f"{state.partnumber('pn-rev')}-feature_tree.py": "feature tree",
        f"{state.partnumber('pn-rev')}-instances_list.tsv": "instances list",
        f"{state.partnumber('pn-rev')}-library_import_history.tsv": "library history",
        "instance_data": {},
        "features_for_relatives": {},
        "harnesses": {},
        "lists": {
            f"{state.partnumber('pn-rev')}-bom.tsv": "bom",
            f"{state.partnumber('pn-rev')}-circuits_list.tsv": "circuits list",
            f"{state.partnumber('pn-rev')}-post_harness_instances_list.tsv": "post harness instances list",
            f"{state.partnumber('pn-rev')}-harness_manifest.tsv": "harness manifest",
            f"{state.partnumber('pn-rev')}-system_connector_list.tsv": "system connector list",
            f"{state.partnumber('pn-rev')}-mapped_channels_set.tsv": "mapped channels set",
            f"{state.partnumber('pn-rev')}-mapped_disconnect_channels_set.tsv": "mapped disconnects set",
            f"{state.partnumber('pn-rev')}-mapped_a_channels_through_disconnects_set.tsv": "mapped A-side channels through disconnects set",
        },
        "maps": {
            f"{state.partnumber('pn-rev')}-channel_map.tsv": "channel map",
            f"{state.partnumber('pn-rev')}-disconnect_map.tsv": "disconnect map",
        },
    }


def generate_structure():
    os.makedirs(fileio.dirpath("instance_data"), exist_ok=True)
    os.makedirs(fileio.dirpath("features_for_relatives"), exist_ok=True)
    os.makedirs(fileio.dirpath("harnesses"), exist_ok=True)
    os.makedirs(fileio.dirpath("lists"), exist_ok=True)
    os.makedirs(fileio.dirpath("maps"), exist_ok=True)
    pass


def render():
    state.set_net(None)

    if not os.path.exists(fileio.path("feature tree")):
        with open(fileio.path("feature tree"), "w", encoding="utf-8") as f:
            f.write(system_feature_tree_utils_default)

    library_history.new()
    instances_list.new()
    cli.print_import_status_headers()
    runpy.run_path(fileio.path("feature tree"))

    post_harness_instances_list.rebuild()

    print("\nSystem rendered successfully!\n")



import os
import os.path
import datetime
import shutil
import re
import csv
import json
import subprocess
from harnice import state

# standard punctuation:
#  .  separates between name hierarchy levels
#  _  means nothing, basically a space character
#  -  if multiple instances are found at the same hierarchy level with the same name,
# this separates name from unique instance identifier


def part_directory():
    return os.path.dirname(os.getcwd())


def rev_directory():
    return os.getcwd()


def silentremove(filepath):
    """
    Removes a file or directory and its contents.

    Args:
        filepath (str): The path to the file or directory to remove.
    """
    if os.path.exists(filepath):
        if os.path.isfile(filepath) or os.path.islink(filepath):
            os.remove(filepath)  # remove file or symlink
        elif os.path.isdir(filepath):
            shutil.rmtree(filepath)  # remove directory and contents


def path(target_value, structure_dict=None, base_directory=None):

    # returns the filepath/filename of a filekey.
    """
    Recursively searches for a value in a nested JSON structure and returns the path to the element containing that value.

    Args:
        target_value (str): The value to search for.

    Returns:
        list: A list of container names leading to the element containing the target value, or None if not found.
    """

    # FILES NOT DEPENDENT ON PRODUCT TYPE
    if target_value == "revision history":
        file_path = os.path.join(
            part_directory(), f"{state.partnumber('pn')}-revision_history.tsv"
        )
        return file_path

    # FILES DEPENDENT ON HARNICE ROOT

    if target_value == "library locations":
        import harnice

        harnice_root = os.path.dirname(
            os.path.dirname(os.path.dirname(harnice.__file__))
        )
        return os.path.join(harnice_root, "library_locations.csv")

    if target_value == "project locations":
        import harnice

        harnice_root = os.path.dirname(
            os.path.dirname(os.path.dirname(harnice.__file__))
        )
        return os.path.join(harnice_root, "project_locations.csv")

    if target_value == "drawnby":
        import harnice

        harnice_root = os.path.dirname(
            os.path.dirname(os.path.dirname(harnice.__file__))
        )
        return os.path.join(harnice_root, "drawnby.json")

    # FILES INSIDE OF A STRUCURE DEFINED BY FILEIO
    # look up from default structure state if not provided
    if structure_dict is None:
        structure_dict = state.file_structure

    def recursive_search(data, path):
        if isinstance(data, dict):
            for key, value in data.items():
                if value == target_value:
                    return path + [key]
                result = recursive_search(value, path + [key])
                if result:
                    return result
        elif isinstance(data, list):
            for index, item in enumerate(data):
                if item == target_value:
                    return path + [f"[{index}]"]
                result = recursive_search(item, path + [f"[{index}]"])
                if result:
                    return result
        return None

    path_value = recursive_search(structure_dict, [])

    if not path_value:
        raise TypeError(f"Could not find filepath of '{target_value}'.")
    if base_directory in [None, ""]:
        return os.path.join(rev_directory(), *path_value)
    else:
        return os.path.join(rev_directory(), base_directory, *path_value)


def dirpath(target_key, structure_dict=None, base_directory=None):
    """
    Returns the absolute path to a directory identified by its key
    within a dict hierarchy.
    """
    if target_key is None:
        if base_directory in [None, ""]:
            return os.path.join(rev_directory())
        else:
            return os.path.join(rev_directory(), base_directory)

    if structure_dict is None:
        structure_dict = state.file_structure

    def recursive_search(data, path):
        if isinstance(data, dict):
            for key, value in data.items():
                # if the current key matches, return its path immediately
                if key == target_key:
                    return path + [key]
                # otherwise, keep descending
                result = recursive_search(value, path + [key])
                if result:
                    return result
        elif isinstance(data, list):
            for index, item in enumerate(data):
                result = recursive_search(item, path + [f"[{index}]"])
                if result:
                    return result
        return None

    path_key = recursive_search(structure_dict, [])
    if not path_key:
        raise TypeError(f"Could not find directory '{target_key}'.")
    if base_directory in [None, ""]:
        return os.path.join(rev_directory(), *path_key)
    else:
        return os.path.join(rev_directory(), base_directory, *path_key)


def verify_revision_structure():
    from harnice import cli
    from harnice.lists import rev_history

    cwd = os.getcwd()
    cwd_name = os.path.basename(cwd)
    parent = os.path.basename(os.path.dirname(cwd))

    # --- 1) Already in a <PN>-revN folder? ---
    if cwd_name.startswith(f"{parent}-rev") and cwd_name.split("-rev")[-1].isdigit():
        state.set_pn(parent)
        state.set_rev(int(cwd_name.split("-rev")[-1]))

    # --- 2) In a part folder that contains revision folders? ---
    elif any(
        re.fullmatch(rf"{re.escape(cwd_name)}-rev\d+", d) for d in os.listdir(cwd)
    ):
        print(f"This is a part folder ({cwd_name}).")
        print(f"Please `cd` into a revision folder (e.g. `{cwd_name}-rev1`) and rerun.")
        exit()

    # --- 3) No revision structure → initialize new PN here ---
    else:
        answer = cli.prompt(
            f"No valid Harnice file structure detected in '{cwd_name}'. Create new PN here?",
            default="y",
        )
        if answer.lower() not in ("y", "yes", ""):
            exit()

        state.set_pn(cwd_name)

        # inline prompt_new_rev
        rev = int(cli.prompt("Enter revision number", default="1"))
        state.set_rev(rev)
        folder = os.path.join(cwd, f"{state.pn}-rev{state.rev}")
        os.makedirs(folder, exist_ok=True)
        os.chdir(folder)

    # --- Ensure revision_history entry exists ---
    try:
        rev_history.info()
    except ValueError:
        rev_history.append(next_rev=state.rev)
    except FileNotFoundError:
        rev_history.append(next_rev=state.rev)

    # --- Status must be blank to proceed ---
    if rev_history.info(field="status") != "":
        raise RuntimeError(
            f"Revision {state.rev} status is not clear. "
            f"Harnice only renders revisions with a blank status."
        )

    print(f"Rendering PN: {state.pn}, Rev: {state.rev}")
    rev_history.update_datemodified()


def today():
    d = datetime.date.today()
    return f"{d.month}/{d.day}/{d.year % 100}"


def get_git_hash_of_harnice_src():
    try:
        # get path to harnice package directory
        import harnice

        repo_dir = os.path.dirname(os.path.dirname(harnice.__file__))
        # ask git for commit hash
        return (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=repo_dir)
            .decode("utf-8")
            .strip()
        )
    except Exception:
        return "UNKNOWN"


def get_path_to_project(traceable_key):
    """
    Given a traceable identifier for a project (PN, URL, etc),
    return the expanded local filesystem path.

    Expects a CSV at the root of the repo named:
        project_locations.csv

    Format (no headers):
        traceable_key,local_path
    """
    from harnice import fileio

    path = fileio.path("project locations")  # resolves to project_locations.csv

    if not os.path.exists(path):
        raise FileNotFoundError(
            "Make a CSV at the root of your Harnice repo called project_locations.csv "
            "with the following format (no headers):\n\n"
            "    traceable_key,local_path\n"
        )

    traceable_key = traceable_key.strip()

    with open(path, newline="", encoding="utf-8") as f:
        reader = csv.reader(f, delimiter=",")
        for row in reader:
            # skip blank or comment lines
            if not row or len(row) < 2 or row[0].strip().startswith("#"):
                continue

            key, local = row[0].strip(), row[1].strip()

            if key == traceable_key:
                if not local:
                    raise ValueError(
                        f"No project local path found for '{traceable_key}'"
                    )
                return os.path.expanduser(local)

    raise ValueError(f"Could not find project traceable key '{traceable_key}'")


def read_tsv(filepath, delimiter="\t"):
    try:
        with open(filepath, newline="", encoding="utf-8") as f:
            return list(csv.DictReader(f, delimiter=delimiter))
    except FileNotFoundError:
        filepath = path(filepath)
        try:
            with open(filepath, newline="", encoding="utf-8") as f:
                return list(csv.DictReader(f, delimiter=delimiter))
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Expected csv or tsv file with delimiter '{delimiter}' at path or key {filepath}"
            )


def drawnby():
    return json.load(open(path("drawnby")))



import os
import ast
from harnice import fileio
from harnice.utils import library_utils


def file_structure():
    return {}


def generate_structure():
    pass


def path(channel_type):
    """
    Args:
        channel_type: tuple like (chid, lib_repo) or string like "(5, '...')"
    """
    chid, lib_repo = parse(channel_type)
    base_dir = library_utils.get_local_path(lib_repo)
    return os.path.join(base_dir, "channel_types", "channel_types.tsv")


def parse(val):
    """Convert stored string into a tuple (chid:int, lib_repo:str)."""
    if not val:
        return None
    if isinstance(val, tuple):
        chid, lib_repo = val
    else:
        chid, lib_repo = ast.literal_eval(str(val))
    return (int(chid), str(lib_repo).strip())


def compatibles(channel_type):
    """
    Look up compatible channel_types for the given channel_type.
    Splits the TSV field by commas and parses each entry into (chid, lib_repo).
    """
    channel_type_id, lib_repo = parse(channel_type)
    for row in fileio.read_tsv(path((channel_type_id, lib_repo))):
        if str(channel_type_id) != str(row.get("channel_type_id")):
            continue

        signals_str = row.get("compatible_channel_types", "").strip()
        if not signals_str:
            return []

        values = [v.strip() for v in signals_str.split(";") if v.strip()]
        parsed = []
        for v in values:
            parsed.append(parse(v))
        return parsed

    return []


def is_or_is_compatible_with(channel_type):
    output = []
    output.append(parse(channel_type))
    for compatible in compatibles(channel_type):
        output.append(compatible)
    return output


# search channel_types.tsv
def signals(channel_type):
    chid, lib_repo = parse(channel_type)

    ch_types_tsv_path = os.path.join(
        library_utils.get_local_path(lib_repo), "channel_types", "channel_types.tsv"
    )

    for row in fileio.read_tsv(ch_types_tsv_path):
        if str(row.get("channel_type_id", "")).strip() == str(chid):
            return [
                sig.strip() for sig in row.get("signals", "").split(",") if sig.strip()
            ]
    return []



import ast
from harnice import fileio
from harnice.lists import instances_list, rev_history
from harnice import state

note_counter = 1


def new_note(
    note_type,
    note_text,
    note_number=None,
    bubble_text=None,
    shape_mpn=None,
    shape_lib_subpath=None,
    shape_lib_repo="https://github.com/harnice/harnice",
    affectedinstances=None,
):
    """
    Creates or updates a note instance.

    Behavior:
    - If a note with identical (note_type, note_text) already exists:
        * If either existing or new affectedinstances is empty → ERROR.
        * Else → merge affectedinstances into existing note, DO NOT create a new one.
    """

    if not shape_mpn:
        shape_mpn = note_type

    global note_counter

    # ------------------------------------------------------------
    # 1. Search for an existing identical note
    # ------------------------------------------------------------
    existing = None
    for instance in fileio.read_tsv("instances list"):
        if (
            instance.get("item_type") == "note"
            and instance.get("note_type") == note_type
            and instance.get("note_text") == note_text
        ):
            existing = instance
            break

    # Normalize new affectedinstances
    new_affected_instances = affectedinstances or []
    if isinstance(new_affected_instances, str):
        import ast

        try:
            new_affected_instances = ast.literal_eval(new_affected_instances)
        except Exception:
            raise ValueError(f"Malformed affectedinstances value: {affectedinstances}")

    # ------------------------------------------------------------
    # 2. CASE A — Existing note found
    # ------------------------------------------------------------
    if existing:
        # Parse existing list safely
        old_affected_instances_raw = existing.get("note_affected_instances")
        if isinstance(old_affected_instances_raw, str):
            import ast

            try:
                old_affected_instances = ast.literal_eval(old_affected_instances_raw)
            except Exception:
                old_affected_instances = []
        else:
            old_affected_instances = old_affected_instances_raw or []

        # A.1 — If either side lacks affected instances → ERROR
        if not old_affected_instances or not new_affected_instances:
            raise ValueError(
                f"Note '{note_type}:{note_text}' already exists but "
                "one of the notes has no affected_instances. "
                "This note has already been assigned."
            )

        # A.2 — Merge and deduplicate into existing instance
        merged = list(
            dict.fromkeys(old_affected_instances + new_affected_instances)
        )  # keep order, remove dupes
        existing["note_affected_instances"] = merged

        # Write modification back to instances_list
        instances_list.modify(
            existing.get("instance_name"), {"note_affected_instances": merged}
        )
        return existing

    # ------------------------------------------------------------
    # 3. CASE B — No existing note, create a new one
    # ------------------------------------------------------------
    instances_list.new_instance(
        f"note-{note_counter}",
        {
            "item_type": "note",
            "note_type": note_type,
            "print_name": bubble_text,
            "note_text": note_text,
            "mpn": shape_mpn,
            "lib_repo": shape_lib_repo,
            "lib_subpath": shape_lib_subpath,
            "note_number": note_number,
            "note_affected_instances": new_affected_instances,
        },
    )

    note_counter += 1


def assign_buildnote_numbers():
    """
    Assigns sequential numbers to build note instances.

    Iterates through all note instances in the instances list and assigns sequential
    numbers to build notes. Each build note gets a unique number and that number is
    set as both the `note_number` and `print_name` fields.
    """
    build_note_counter = 0
    for instance in fileio.read_tsv("instances list"):
        if instance.get("note_type") == "build_note":
            if instance.get("item_type") == "note":
                build_note_counter += 1
            instances_list.modify(
                instance.get("instance_name"),
                {"note_number": build_note_counter, "print_name": build_note_counter},
            )


def make_rev_history_notes(rev):
    """
    Creates revision change callout notes based on revision history.

    Creates a revision change callout note for a given revision, linking it to
    the instances affected by that revision. The note text comes from the
    revision's update description.

    **Args:**
    - `rev` (dict): Revision dictionary from revision history containing at least
        `'rev'` and `'revisionupdates'` fields.
    """
    affected_instances = rev_history.info(rev=rev.get("rev"), field="affectedinstances")

    if affected_instances:  # safer + more pythonic
        new_note(
            note_type="rev_change_callout",
            note_text=rev.get("revisionupdates"),
            note_number=rev.get("rev"),
            bubble_text=rev.get("rev"),
            shape_mpn="rev_change_callout",
            shape_lib_repo="https://github.com/harnice/harnice",
            affectedinstances=affected_instances,
        )


def make_bom_flagnote(affected_instance, output_csys_name):
    """
    Creates a BOM item flagnote dictionary for an instance.

    Creates a flagnote configuration dictionary that displays the BOM line number
    for a given instance. The flagnote is positioned at the specified output
    coordinate system of the instance.

    **Args:**
    - `affected_instance` (dict): Instance dictionary to create a flagnote for.
    - `output_csys_name` (str): Name of the output coordinate system where the
        flagnote should be positioned.

    **Returns:**
    - `dict`: A flagnote instance dictionary ready to be added to the instances list.
    """
    return {
        "net": state.net,
        "instance_name": f"note-bom_item-{affected_instance.get('instance_name')}",
        "print_name": affected_instance.get("bom_line_number"),
        "mpn": "bom_item",
        "item_type": "flagnote",
        "parent_instance": affected_instance.get("instance_name"),
        "segment_group": affected_instance.get("segment_group"),
        "connector_group": affected_instance.get("connector_group"),
        "parent_csys_instance_name": affected_instance.get("instance_name"),
        "parent_csys_outputcsys_name": output_csys_name,
        "absolute_rotation": 0,
        "note_type": "bom_item",
        "note_affected_instances": [affected_instance.get("instance_name")],
        "lib_repo": "https://github.com/harnice/harnice",
    }


def make_part_name_flagnote(affected_instance, output_csys_name):
    """
    Creates a part name flagnote dictionary for an instance.

    Creates a flagnote configuration dictionary that displays the print name
    of a given instance. The flagnote is positioned at the specified output
    coordinate system of the instance.

    **Args:**
    - `affected_instance` (dict): Instance dictionary to create a flagnote for.
    - `output_csys_name` (str): Name of the output coordinate system where the
        flagnote should be positioned.

    **Returns:**
    - `dict`: A flagnote instance dictionary ready to be added to the instances list.
    """
    return {
        "net": state.net,
        "instance_name": f"note-part_name-{affected_instance.get('instance_name')}",
        "print_name": affected_instance.get("print_name"),
        "mpn": "part_name",
        "item_type": "flagnote",
        "parent_instance": affected_instance.get("instance_name"),
        "segment_group": affected_instance.get("segment_group"),
        "connector_group": affected_instance.get("connector_group"),
        "parent_csys_instance_name": affected_instance.get("instance_name"),
        "parent_csys_outputcsys_name": output_csys_name,
        "absolute_rotation": 0,
        "note_type": "part_name",
        "note_affected_instances": [affected_instance.get("instance_name")],
        "lib_repo": "https://github.com/harnice/harnice",
    }


def make_buildnote_flagnote(note_instance, affected_instance, output_csys_name):
    """
    Creates a build note flagnote dictionary linking a note to an instance.

    Creates a flagnote configuration dictionary that displays a build note on
    a specific instance. The flagnote shows the build note's print name and is
    positioned at the specified output coordinate system of the affected instance.

    **Args:**
    - `note_instance` (dict): Build note instance dictionary.
    - `affected_instance` (dict): Instance dictionary to attach the flagnote to.
    - `output_csys_name` (str): Name of the output coordinate system where the
        flagnote should be positioned.

    **Returns:**
    - `dict`: A flagnote instance dictionary ready to be added to the instances list.
    """
    return {
        "net": state.net,
        "instance_name": f"note-build_note-{note_instance.get('instance_name')}-{affected_instance.get('instance_name')}",
        "print_name": note_instance.get("print_name"),
        "mpn": "build_note",
        "item_type": "flagnote",
        "parent_instance": affected_instance.get("instance_name"),
        "segment_group": affected_instance.get("segment_group"),
        "connector_group": affected_instance.get("connector_group"),
        "parent_csys_instance_name": affected_instance.get("instance_name"),
        "parent_csys_outputcsys_name": output_csys_name,
        "absolute_rotation": 0,
        "note_type": "build_note",
        "note_affected_instances": [affected_instance.get("instance_name")],
        "lib_repo": "https://github.com/harnice/harnice",
    }


def make_rev_change_flagnote(note_instance, affected_instance, output_csys_name):
    """
    Creates a revision change callout flagnote dictionary linking a note to an instance.

    Creates a flagnote configuration dictionary that displays a revision change
    callout note on a specific instance. The flagnote shows the revision number
    and is positioned at the specified output coordinate system of the affected instance.

    **Args:**
    - `note_instance` (dict): Revision change callout note instance dictionary.
    - `affected_instance` (dict): Instance dictionary to attach the flagnote to.
    - `output_csys_name` (str): Name of the output coordinate system where the
        flagnote should be positioned.

    **Returns:**
    - `dict`: A flagnote instance dictionary ready to be added to the instances list.
    """
    return {
        "net": state.net,
        "instance_name": f"note-rev_change_callout-{note_instance.get('instance_name')}-{affected_instance.get('instance_name')}",
        "print_name": note_instance.get("print_name"),
        "mpn": "rev_change_callout",
        "item_type": "flagnote",
        "parent_instance": affected_instance.get("instance_name"),
        "segment_group": affected_instance.get("segment_group"),
        "connector_group": affected_instance.get("connector_group"),
        "parent_csys_instance_name": affected_instance.get("instance_name"),
        "parent_csys_outputcsys_name": output_csys_name,
        "absolute_rotation": 0,
        "note_type": "rev_change_callout",
        "note_affected_instances": [affected_instance.get("instance_name")],
        "lib_repo": "https://github.com/harnice/harnice",
    }


def parse_note_instance(instance):
    """
    Return a full copy of `instance`, but with note_affected_instances
    parsed into a real Python list (or left alone if blank).
    """
    parsed = {}

    for key, value in instance.items():
        if key == "note_affected_instances":
            if isinstance(value, str) and value.strip():
                try:
                    parsed[key] = ast.literal_eval(value)
                except Exception:
                    parsed[key] = []  # fallback if malformed
            else:
                parsed[key] = []
        else:
            parsed[key] = value

    return parsed


def get_lib_build_notes(instance):
    """
    Returns list of build_notes for this instance from the TSV row.
    Safely parses with ast.literal_eval.
    Always returns a Python list.
    """

    raw = instance.get("lib_build_notes")

    if not raw or raw in ["", None]:
        return []

    try:
        # Expecting a string representation of a list
        parsed = ast.literal_eval(raw)

        # Ensure it's actually a list
        return parsed if isinstance(parsed, list) else []

    except Exception:
        # Malformed literal → fail safe
        return []


def get_lib_tools(instance):
    """
    Returns list of tools for this instance from the TSV row.
    Safely parses with ast.literal_eval.
    Always returns a Python list.
    """

    raw = instance.get("lib_tools")

    if not raw or raw in ["", None]:
        return []

    try:
        # Expecting a string representation of a list
        parsed = ast.literal_eval(raw)

        # Ensure it's actually a list
        return parsed if isinstance(parsed, list) else []

    except Exception:
        # Malformed literal → fail safe
        return []


def combine_notes(keep_note_text, merge_note_texts, note_type=None):
    """
    Combines multiple notes by merging their affected instances into one note.

    Merges one or more notes into a single note by combining their affected instances
    lists. The note to keep is identified by `keep_note_text`, and all notes matching
    `merge_note_texts` are merged into it and then removed.

    **Args:**
    - `keep_note_text` (str): The `note_text` value of the note to keep and merge others into.
    - `merge_note_texts` (list): List of `note_text` values to find and merge into the kept note.
    - `note_type` (list, optional): If provided, only notes with `note_type` in this list
        will be considered for merging. If `None`, all notes are considered.
    """
    keep_note_instance = None
    merge_note_instances_raw = []
    for instance in fileio.read_tsv("instances list"):
        if instance.get("note_text") in [None, ""]:
            continue
        if instance.get("note_text") == keep_note_text:
            if note_type:
                if instance.get("note_type") in note_type:
                    keep_note_instance = instance
            else:
                keep_note_instance = instance
        if instance.get("note_text") in merge_note_texts:
            if note_type:
                if instance.get("note_type") in note_type:
                    merge_note_instances_raw.append(instance)
            else:
                merge_note_instances_raw.append(instance)

    merge_note_instances_parsed = []
    for instance in merge_note_instances_raw:
        merge_note_instances_parsed.append(parse_note_instance(instance))

    new_affected_instances = set()

    for merged_note_instance in merge_note_instances_parsed:
        for affected_instance in merged_note_instance.get("note_affected_instances"):
            new_affected_instances.add(affected_instance)
        instances_list.remove_instance(merged_note_instance)

    for affected_instance in parse_note_instance(keep_note_instance).get(
        "note_affected_instances"
    ):
        new_affected_instances.add(affected_instance)

    instances_list.modify(
        keep_note_instance.get("instance_name"),
        {"note_affected_instances": list(new_affected_instances)},
    )
